
__author__    = "Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import sys
import time

from PIL import Image

import radica.saga as rs


REMOTE_JOB_ENDPOINT = "condor://localhost"

# the dimension (in pixel) of the whole fractal
IMGX = 2048
IMGY = 2048

# the number of tiles in X and Y direction
TILESX = 2
TILESY = 2

if __name__ == "__main__":
    try:
        # list that holds the jobs
        jobs = []

        jobservice = rs.job.Service(REMOTE_JOB_ENDPOINT)

        for x in range(0, TILESX):
            for y in range(0, TILESY):

                # describe a single Mandelbrot job. we're using the
                # directory created above as the job's working directory
                outputfile = 'tile_x%s_y%s.gif' % (x, y)
                jd = rs.job.Description()
                # candidate hosts can be changed / and or commented out
                # the list below seems to be a good working set for OSG
                #jd.candidate_hosts = ["FNAL_FERMIGRID", "cinvestav", "SPRACE",
                #                      "NYSGRID_CORNELL_NYS1", "Purdue-Steele",
                #                      "MIT_CMS_CE2", "SWT2_CPB", "AGLT2_CE_2",
                #                      "UTA_SWT2", "GridUNESP_CENTRAL",
                #                      "USCMS-FNAL-WC1-CE3"]
                # on OSG we need to stage in the data with the jobs. we
                # can't use the saga filesystem API to copy data around since
                # the execution location of the jobs is not known a priori
                jd.file_transfer     = ["mandelbrot.sh > mandelbrot.sh",
                                        "mandelbrot.py > mandelbrot.py",
                                        "%s < %s" % (outputfile, outputfile)]
                jd.wall_time_limit   = 10
                jd.total_cpu_count   = 1
                jd.executable        = '/bin/sh'
                jd.arguments         = ['mandelbrot.sh', IMGX, IMGY, 
                                        (IMGX/TILESX*x), (IMGX/TILESX*(x+1)),
                                        (IMGY/TILESY*y), (IMGY/TILESY*(y+1)),
                                        outputfile]
                # create the job from the description
                # above, launch it and add it to the list of jobs
                job = jobservice.create_job(jd)
                job.run()
                jobs.append(job)
                print(' * Submitted %s. Output will be written to: %s' % (job.id, outputfile))

        # wait for all jobs to finish
        while len(jobs) > 0:
            for job in jobs:
                jobstate = job.get_state()
                print(' * Job %s status: %s' % (job.id, jobstate))
                if jobstate in [rs.job.DONE, rs.job.FAILED]:
                    jobs.remove(job)
            time.sleep(5)

        # stitch together the final image
        fullimage = Image.new('RGB', (IMGX, IMGY), (255, 255, 255))
        print(' * Stitching together the whole fractal: mandelbrot_full.gif')
        for x in range(0, TILESX):
            for y in range(0, TILESY):
                partimage = Image.open('tile_x%s_y%s.gif' % (x, y))
                fullimage.paste(partimage,
                                (IMGX/TILESX*x, IMGY/TILESY*y,
                                 IMGX/TILESX*(x+1), IMGY/TILESY*(y+1)))
        fullimage.save("mandelbrot_full.gif", "GIF")
        sys.exit(0)

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occured: (%s) %s " % (ex.type, (str(ex))))
        # Trace back the exception. That can be helpful for debugging.
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        sys.exit(-1)

    except KeyboardInterrupt:
    # ctrl-c caught: try to cancel our jobs before we exit
        # the program, otherwise we'll end up with lingering jobs.
        for job in jobs:
            job.cancel()
        sys.exit(-1)




__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import decorators as cpi_dec
from .. import namespace  as cpi_ns

SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL

# ------------------------------------------------------------------------------
# keep order of inheritance!  super() below uses MRO
class Directory (cpi_ns.directory.Directory) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsdirec = super  (Directory, self)
        self._cpi_nsdirec.__init__ (api, adaptor)

    @SYNC
    def init_instance         (self, url, flags, session) : pass
    @ASYNC
    def init_instance_async   (self, url, flags, session) : pass


    # ----------------------------------------------------------------
    #
    # add filesystem directory methods
    #
    @SYNC
    def get_size              (self, name, flags, ttype)  : pass
    @ASYNC
    def get_size_async        (self, name, flags, ttype)  : pass

    @SYNC
    def get_size_self         (self,       flags, ttype)  : pass
    @ASYNC                                
    def get_size__self_async  (self,       flags, ttype)  : pass

    @SYNC
    def is_file               (self, name,        ttype)  : pass
    @ASYNC
    def is_file_async         (self, name,        ttype)  : pass

    @SYNC
    def is_file_self          (self,              ttype)  : pass
    @ASYNC                                        
    def is_file_self_async    (self,              ttype)  : pass


# ------------------------------------------------------------------------------




#!/usr/bin/env python

""" This examples shows how to run a job on LoadLeveler cluster.

    More information about the radical.saga job API can be found at:
    http://radical-cybertools.github.com/radical.saga/doc/library/job/index.html
"""

__author__    = "Ole Weidner"
__copyright__ = "Copyright 2013, The SAGA Project"
__license__   = "MIT"

import sys

import radical.saga as rs


CLUSTER_HOST = "your.loadleveler.cluster.hostname"
CLUSTER_NAME = "your.loadleveler.cluster.name"

# ----------------------------------------------------------------------------
#
def main():

    try:
        # Your ssh identity on the remote machine
        ctx = rs.Context("ssh")
        ctx.user_id = "your_username"

        session = rs.Session()
        session.add_context(ctx)

        # Create a job service object that represent a remote loadleveler
        # cluster. The keyword 'loadl' in the url scheme triggers the
        # LoadLeveler adaptors and '+ssh' enables LoadLeveler remote access
        # via SSH. and 'cluster' URL query specify loadleveler cluster name.
        # "llq -X cluster"
        js = rs.job.Service("loadl+ssh://%s?cluster=%s" % \
            (CLUSTER_HOST, CLUSTER_NAME), session=session)

        # describe our job
        jd = rs.job.Description()

        # Next, we describe the job we want to run. A complete set of job
        # description attributes can be found in the API documentation.
        jd.environment     = {'MYOUTPUT':'"Hello LoadLevler Adaptor from SAGA"'}
        jd.executable      = '/bin/echo'
        jd.arguments       = ['$MYOUTPUT']
        jd.output          = "/tmp/mysagajob.stdout"
        jd.error           = "/tmp/mysagajob.stderr"

        # Create a new job from the job description. The initial state of
        # the job is 'New'.
        myjob = js.create_job(jd)

        # Check our job's id and state
        print("Job ID    : %s" % (myjob.id))
        print("Job State : %s" % (myjob.state))

        print("\n...starting job...\n")

        # Now we can start our job.
        myjob.run()

        print("Job ID    : %s" % (myjob.id))
        print("Job State : %s" % (myjob.state))

        print("\n...waiting for job...\n")
        # wait for the job to either finish or fail
        myjob.wait()

        print("Job ID    : %s" % (myjob.id))
        print("Job State : %s" % (myjob.state))
        print("Exitcode  : %s" % (myjob.exit_code))

        js.close()
        return 0

    except rs.SagaException as ex:
        # Catch all saga exceptions
        print("An exception occurred: (%s) %s " % (ex.type, (str(ex))))
        # Get the whole traceback in case of an exception -
        # this can be helpful for debugging the problem
        print(" \n*** Backtrace:\n %s" % ex.traceback)
        return -1

if __name__ == "__main__":
    sys.exit(main())




__author__    = "Andre Merzky, Ashley Z, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" mock based resource adaptor implementation """

import re
import time

from ...                   import exceptions as rse
from ...                   import url        as rsurl
from ...utils              import pty_mock  as rsups
from ...                   import resource   as api_resource
from ...adaptors           import base       as a_base
from ...adaptors.cpi       import resource   as cpi_resource
from ...adaptors.cpi       import decorators as cpi_decs
from ...resource           import constants  as c


c.ANY = c.COMPUTE | c.STORAGE

SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


# --------------------------------------------------------------------
# the adaptor info
#
_ADAPTOR_NAME          = "radical.saga.adaptors.mock_resource"
_ADAPTOR_SCHEMAS       = ["local", "mock"]
_ADAPTOR_OPTIONS       = []

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "rdes_attributes"  : [c.RTYPE         ,
                          c.MACHINE_OS    ,
                          c.MACHINE_ARCH  ,
                          c.SIZE          ,
                          c.MEMORY        ,
                          c.ACCESS       ],
    "res_attributes"   : [c.RTYPE         ,
                          c.MACHINE_OS    ,
                          c.MACHINE_ARCH  ,
                          c.SIZE          ,
                          c.MEMORY        ,
                          c.ACCESS       ],
    "metrics"          : [c.STATE,
                          c.STATE_DETAIL],
    "contexts"         : {"ssh"      : "public/private keypair",
                          "x509"     : "X509 proxy for gsissh",
                          "userpass" : "username/password pair for ssh"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"             : _ADAPTOR_NAME,
    "cfg_options"      : _ADAPTOR_OPTIONS,
    "capabilities"     : _ADAPTOR_CAPABILITIES,
    "description"      : """
        The Mock resource adaptor. This adaptor attempts to determine what job
        submission endpoint and file system resources are available for a given
        host, and provides the respective access URLs.
        """,
    "example": "examples/jobs/localresource.py",
    "schemas"          : {"mock"  : "find access URLs for mock-job/file adaptors"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO          = {
    "name"             : _ADAPTOR_NAME,
    "version"          : "v0.1.beta",
    "schemas"          : _ADAPTOR_SCHEMAS,
    "cpis"             : [
        {
        "type"         : "radical.saga.resource.Manager",
        "class"        : "MockResourceManager"
        },
        {
        "type"         : "radical.saga.resource.Compute",
        "class"        : "MockResourceCompute"
        },
    ]
}


###############################################################################
# The adaptor class

class Adaptor (a_base.Base):
    """
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """


    # ----------------------------------------------------------------
    #
    def __init__ (self) :

        self.base = super  (Adaptor, self)
        self.base.__init__ (_ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        # for id parsing
        self.id_re = re.compile ('^\[(.*)\]-\[(.*?)\]$')


    # ----------------------------------------------------------------
    #
    def sanity_check (self) :

        # FIXME: also check for gsissh

        pass


    # ----------------------------------------------------------------
    #
    def parse_id (self, id) :
        # split the id '[manager-url]-[resource-url]' in its parts, and return them.

        match = self.id_re.match (id)

        if  not match or len (match.groups()) != 2 :
            raise rse.BadParameter ("Cannot parse resource id '%s'" % id)

        return (rsurl.Url (match.group(1)), rsurl.Url (match.group (2)))




###############################################################################
#
class MockResourceManager (cpi_resource.Manager) :

    # ----------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (MockResourceManager, self)
        self._cpi_base.__init__ (api, adaptor)



    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_state, url, session) :

        self.url     = rsurl.Url (url)  # deep copy
        self.session = session
        self.access  = {}
        self.access[c.COMPUTE] = []
        self.access[c.STORAGE] = []
        self.access[c.ANY]     = []

        # check for compute entry points
        for schema in ['fork', 'ssh', 'gsissh'] :
            tmp_url = rsurl.Url (self.url)  # deep copy
            tmp_url.schema = schema

            mock = rsups.PTYMock (tmp_url, self.session, self._logger)

            if  mock.alive () :
                self.access[c.COMPUTE].append (tmp_url)
                self.access[c.ANY]    .append (tmp_url)
                mock.finalize (True)


        # check for storage entry points
        for schema in ['file', 'sftp', 'gsisftp'] :
            tmp_url = rsurl.Url (self.url)  # deep copy
            tmp_url.schema = schema

            mock = rsups.PTYMock (tmp_url, self.session, self._logger)

            if  mock.alive () :
                self.access[c.STORAGE].append (tmp_url)
                self.access[c.ANY]    .append (tmp_url)
                mock.finalize (True)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def acquire (self, rd) :

        if  not rd :
            raise rse.BadParameter._log (self._logger, "acquire needs a resource description")

        if  rd.rtype != c.COMPUTE and \
            rd.rtype != c.STORAGE     :
            raise rse.BadParameter._log (self._logger, "can only acquire compute and storage resources.")


        # check that only supported attributes are provided
        for attribute in rd.list_attributes():
            if attribute not in _ADAPTOR_CAPABILITIES["rdes_attributes"]:
                msg = "'resource.Description.%s' is not supported by this adaptor" % attribute
                raise rse.BadParameter._log (self._logger, msg)

        if  rd.access :
            access_url = rsurl.Url (rd.access)
            if  access_url not in self.access[rd.rtype] :
                msg = "access '%s' is not supported by this backend" % rd.access
                raise rse.BadParameter._log (self._logger, msg)

        if  not len (self.access[rd.rtype]) :
            raise rse.BadParameter._log (self._logger, "resource type is not supported by this backend")


        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"resource_access_url"     : self.access[rd.rtype][0],
                         "resource_type"           : rd.rtype,
                         "resource_manager"        : self.get_api (),
                         "resource_manager_url"    : self.url,
                         "resource_schema"         : self.url.schema }

        if rd.rtype == c.COMPUTE :
            return api_resource.Compute (_adaptor=self._adaptor, _adaptor_state=adaptor_state)

        if rd.rtype == c.STORAGE :
            return api_resource.Storage (_adaptor=self._adaptor, _adaptor_state=adaptor_state)


    # ----------------------------------------------------------------
    @SYNC_CALL
    def get_url (self) :

        return self.url


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list (self, rtype):

        return self.access[rtype]


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def release (self, id):

        return  # hahaha


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list_templates (self, rtype) :

        return []  # no templates


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_template (self, name) :

        raise rse.BadParameter ("unknown template %s" % name)


###############################################################################
#
class MockResourceCompute (cpi_resource.Compute) :

    # ----------------------------------------------------------------
    #
    def __init__ (self, api, adaptor) :

        self._cpi_base = super  (MockResourceCompute, self)
        self._cpi_base.__init__ (api, adaptor)

        self.state       = c.ACTIVE
        self.rtype       = None
        self.manager     = None
        self.manager_url = None
        self.access_url  = None


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance (self, adaptor_info, id, session):

        # eval id if given
        if  id :
            self.manager_url, self.access_url = self._adaptor.parse_id (id)
            self.manager = api_resource.Manager (self.manager_url)

            if  self.access_url.scheme in self.manager.list (c.COMPUTE) :
                self.rtype = c.COMPUTE

            elif self.access_url.scheme in self.manager.list (c.STORAGE) :
                self.rtype = c.STORAGE

            else :
                raise rse.BadParameter ("Cannot handle resource type for %s", id)

        # no id -- grab info from adaptor_info
        elif adaptor_info :

            if  'resource_access_url'  not in adaptor_info or \
                'resource_type'        not in adaptor_info or \
                'resource_manager'     not in adaptor_info or \
                'resource_manager_url' not in adaptor_info    :
                raise rse.BadParameter ("Cannot acquiure resource, insufficient information")

            self.access_url  = adaptor_info['resource_access_url']
            self.rtype       = adaptor_info['resource_type']
            self.manager     = adaptor_info['resource_manager']
            self.manager_url = adaptor_info['resource_manager_url']

            self.id = "[%s]-[%s]" % (self.manager_url, self.access_url)

        else :
            raise rse.BadParameter ("Cannot acquire resource, no contact information")


        return self.get_api ()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id           (self) : return self.id

    @SYNC_CALL
    def get_rtype        (self) : return self.rtype

    @SYNC_CALL
    def get_state        (self) : return self.state

    @SYNC_CALL
    def get_state_detail (self) : return None

    @SYNC_CALL
    def get_access       (self) : return self.access_url

    @SYNC_CALL
    def get_manager      (self) : return self.manager

    @SYNC_CALL
    def get_description  (self) : return {c.ACCESS : self.access_url}


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def reconfig (self):
        raise rse.NotImplemented("This backend cannot reconfigre resources")


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def release (self):

        return  # hahahah


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def wait (self, state, timeout) :
        # trick is, we *never* change state...

        if  state == self.state :
            return

        if  timeout >= 0 :
            time.sleep (timeout)
            return

        if  timeout < 0 :
            while True :
                time.sleep (10)

        # we never get here...








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


################################################################################
#
# Job States

UNKNOWN               = 'Unknown';   """ The state of the job could not be 
                                         determined.  """ 
NEW                   = 'New';       """ This state identifies a newly 
                                         constructed job instance which has not 
                                         yet been submitted / started to run. 
                                         """ 
PENDING               = 'Pending';   """ This state identifies a job instance 
                                         which has been submitted / started but 
                                         is not yet executing on the resource. 
                                         """ # non-GFD.90
RUNNING               = 'Running';   """ This state identifies a job instance 
                                         which has been submitted and is 
                                         currently running.  """ 
SUSPENDED             = 'Suspended'; """ The job has been suspended by the 
                                         job service.  """
DONE                  = 'Done';      """ The job has finished succesfully.  
                                         This state is final.  """
FAILED                = 'Failed';    """ The job has finished unsuccessfully / 
                                         with an error.  
                                         This state is final. """ 
CANCELED              = 'Canceled';  """ The job has been canceled either by the 
                                         user or by job service.  
                                         This state is final.  """ 

FINAL                 = [DONE, FAILED, CANCELED]

################################################################################
#
# JobDescription attributes:
EXECUTABLE            = 'Executable';          """ The path to the application
                                                   executable """
PRE_EXEC              = 'PreExec';             """ List of commands to run
                                                   before job execution, on the
                                                   target resource """        # non-GFD.90
POST_EXEC             = 'PostExec';            """ List of commands to run
                                                   after successfull job execution,
                                                   on the target resource """ # non-GFD.90
ARGUMENTS             = 'Arguments';           """ The arguments for to the
                                                   executable  """
ENVIRONMENT           = 'Environment';         """ dict, containing environment 
                                                   settings for the job """ 
WORKING_DIRECTORY     = 'WorkingDirectory';    """ :todo: docstring """ 
INTERACTIVE           = 'Interactive';         """ :todo: docstring """ 
INPUT                 = 'Input';               """ :todo: docstring """ 
OUTPUT                = 'Output';              """ :todo: docstring """ 
ERROR                 = 'Error';               """ :todo: docstring """ 
PROJECT               = 'Project';             """ :todo: docstring """ 
FILE_TRANSFER         = 'FileTransfer';        """ :todo: docstring """ 
CLEANUP               = 'Cleanup';             """ :todo: docstring """ 
JOB_START_TIME        = 'JobStartTime';        """ :todo: docstring """ 
WALL_TIME_LIMIT       = 'WallTimeLimit';       """ :todo: docstring """ 
TOTAL_CPU_TIME        = 'TotalCPUTime';        """ :todo: docstring """ 
TOTAL_PHYSICAL_MEMORY = 'TotalPhysicalMemory'; """ :todo: docstring """
SYSTEM_ARCHITECTURE   = 'SystemArchitecture';  """ :todo: docstring """
OPERATING_SYSTEM_TYPE = 'OperatingSystemType'; """ :todo: docstring """ 
CANDIDATE_HOSTS       = 'CandidateHosts';      """ :todo: docstring """ 
QUEUE                 = 'Queue';               """ :todo: docstring """ 
SPMD_VARIATION        = 'SPMDVariation';       """ The type of parallelism 
                                                   required by this job """ 
TOTAL_CPU_COUNT       = 'TotalCPUCount';       """ The number of CPUs required 
                                                   by this job """
TOTAL_GPU_COUNT       = 'TotalGPUCount';       """ The number of GPUs required 
                                                   by this job """
NUMBER_OF_PROCESSES   = 'NumberOfProcesses';   """ Number of instances of """ 
PROCESSES_PER_HOST    = 'ProcessesPerHost';    """ :todo: docstring """ 
THREADS_PER_PROCESS   = 'ThreadsPerProcess';   """ :todo: docstring """ 
JOB_CONTACT           = 'JobContact';          """ :todo: docstring """
NAME                  = 'Name';                """ The name of your job """ # non-GFD.90


################################################################################
# Job attributes:
ID                    = 'ID';             """ :todo: docstring """ 
EXECUTION_HOSTS       = 'ExecutionHosts'; """ :todo: docstring """ 
CREATED               = 'Created';        """ :todo: docstring """ 
STARTED               = 'Started';        """ :todo: docstring """ 
FINISHED              = 'Finished';       """ :todo: docstring """ 
EXIT_CODE             = 'ExitCode';       """ :todo: docstring """ 
TERMSIG               = 'Termsig';        """ :todo: docstring """ 
SERVICE_URL           = 'ServiceUrl';     """ :todo: docstring """ # non-GFD.90

################################################################################
# Job metrics:
STATE                 = 'State';       """ Subscribable job state information. 
                                           This metric gets triggered whenever 
                                           the state of the job changes.  """ 
STATE_DETAIL          = 'StateDetail'; """ Allows to get information about the
                                           native (backend) job state. For some 
                                           applications, access to the native 
                                           backend state model can be important,
                                           however, it is not guaranteed to be 
                                           supported by all middleware adaptors.
                                           Generally, state details is supposed 
                                           to be formatted as follows::

                                             '<model>:<state>'=value
                                       
                                           The STATE_DETAIL metric gets triggered 
                                           whenever the backend state information 
                                           of the job changes.  """
SIGNAL                = 'Signal';      """ :todo: docstring """ 
CPU_TIME              = 'CPUTime';     """ :todo: docstring """ 
MEMORY_USE            = 'MemoryUse';   """ :todo: docstring """ 
VMEMORY_USE           = 'VmemoryUse';  """ :todo: docstring """ 
PERFORMANCE           = 'Performance'; """ :todo: docstring """ 








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import sys
import inspect

import radical.utils              as ru
import radical.utils.signatures   as rus

from .adaptors import base        as sab
from .         import engine


# ------------------------------------------------------------------------------
#
class SimpleBase (object) :
    """ This is a very simple API base class which just initializes
    the self._logger and self._engine members, but does not perform any further
    initialization, nor any adaptor binding.  This base is used for API classes
    which are not backed by multiple adaptors (no session, tasks, etc).
    """

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('SimpleBase')
    @rus.returns (rus.nothing)
    def __init__  (self, uid=None):

        ok = False
        try:
            ok = hasattr(self, '_apitype')

        except:
            pass

        if not ok:
            self._apitype = self._get_apitype()


        self._logger = ru.Logger('radical.saga')
        if uid:
            self._id = uid
        else:
            self._id = ru.generate_id (self._apitype, mode=ru.ID_SIMPLE)

      # self._logger.debug ("[saga.Base] %s.__init__()" % self._apitype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('SimpleBase')
    @rus.returns (str)
    def _get_apitype (self) :

        # apitype for saga.job.service.Service should be saga.job.Service --
        # but we need to make sure that this actually exists and is equivalent.

        mname_1 = self.__module__            # saga.job.service
        cname   = self.__class__.__name__    # Service

        mname_1_elems = mname_1.split ('.')  # ['saga', 'job', 'service']
        mname_1_elems.pop ()                 # ['saga', 'job']
        mname_2 = '.'.join (mname_1_elems)   # 'saga.job'

        for mod_cname, mod_class in inspect.getmembers (sys.modules[mname_2]) :
            if  mod_cname == cname           and \
                inspect.isclass (mod_class)  and \
                isinstance (self, mod_class)     :

                apitype = "%s.%s" % (mname_2, cname)  # saga.job.Service
                return apitype

        apitype = "%s.%s" % (mname_1, cname)  # saga.job.service.Service
        return apitype


# ------------------------------------------------------------------------------
#
class Base (SimpleBase) :


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Base',
                  str,
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.anything),
                  rus.optional (rus.anything))
    @rus.returns (rus.nothing)
    def __init__  (self, schema, adaptor, adaptor_state, *args, **kwargs) :

        SimpleBase.__init__ (self)

        _engine       = engine.Engine ()
        self._adaptor = adaptor
        self._adaptor = _engine.bind_adaptor(self, self._apitype, schema,
                                             adaptor)

        # Sync creation (normal __init__) will simply call the adaptor's
        # init_instance at this point.  _init_task should *not* be evaluated,
        # ever, for __init__ based construction! (it is private, see?)
        #
        # For any async creation (create()), we need to return a task which
        # performs initialization.  We rely on the sync/async method decorators
        # on CPI level to provide the task instance itself, and point the task's
        # workload to the adaptor level init_instance method.

        self._init_task = self._adaptor.init_instance(adaptor_state, *args,
                                                      **kwargs)

        if 'ttype' in kwargs and kwargs['ttype'] :
            # in this case we in in fact need the init_task later on, to return
            # on the create() call
            pass
        else :
            # in the sync case, we can get rid of the init_task reference, to
            # simplify garbage collection
            self._init_task = None


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Base')
    @rus.returns ('saga.Session')
    def get_session (self) :
        """
        Returns the session which is managing the object instance.  For objects
        which do not accept a session handle on construction, this call returns
        None.

        The object's session is also available via the `session` property.
        """
        return self._adaptor.get_session ()

    session = property (get_session)


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" Provides the SAGA Job CPI """

from ..decorators import CPI_SYNC_CALL  as SYNC
from ..decorators import CPI_ASYNC_CALL as ASYNC
from ..base       import CPIBase
from ..sasync     import Async


class Job (CPIBase, Async) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        _cpi_base = super  (Job, self)
        _cpi_base.__init__ (api, adaptor)

    @SYNC
    def init_instance             (self, info, ttype)    : pass
    @ASYNC
    def init_instance_async       (self, info, ttype)    : pass


    # ----------------------------------------------------------------
    #
    # job methods
    #
    @SYNC
    def get_id                    (self, ttype)          : pass
    @ASYNC
    def get_id_async              (self, ttype)          : pass

    @SYNC
    def get_name                  (self, ttype)          : pass
    @ASYNC
    def get_name_async            (self, ttype)          : pass

    @SYNC
    def get_description           (self, ttype)          : pass
    @ASYNC
    def get_description_async     (self, ttype)          : pass

    @SYNC
    def get_stdin                 (self, ttype)          : pass
    @ASYNC
    def get_stdin_async           (self, ttype)          : pass

    @SYNC
    def get_stdout                (self, ttype)          : pass
    @ASYNC
    def get_stdout_async          (self, ttype)          : pass

    @SYNC
    def get_stderr                (self, ttype)          : pass
    @ASYNC
    def get_stderr_async          (self, ttype)          : pass

    @SYNC
    def get_log                   (self, ttype)          : pass
    @ASYNC
    def get_log_async             (self, ttype)          : pass

    @SYNC
    def suspend                   (self, ttype)          : pass
    @ASYNC
    def suspend_async             (self, ttype)          : pass

    @SYNC
    def resume                    (self, ttype)          : pass
    @ASYNC
    def resume_async              (self, ttype)          : pass

    @SYNC
    def checkpoint                (self, ttype)          : pass
    @ASYNC
    def checkpoint_async          (self, ttype)          : pass

    @SYNC
    def migrate                   (self, jd, ttype)      : pass
    @ASYNC
    def migrate_async             (self, jd, ttype)      : pass

    @SYNC
    def signal                    (self, signum, ttype)  : pass
    @ASYNC
    def signal_async              (self, signum, ttype)  : pass


    #-----------------------------------------------------------------
    # task methods flattened into job :-/
    @SYNC
    def run                       (self, ttype)          : pass
    @ASYNC
    def run_async                 (self, ttype)          : pass

    @SYNC
    def cancel                    (self, timeout, ttype) : pass
    @ASYNC
    def cancel_async              (self, timeout, ttype) : pass

    @SYNC
    def wait                      (self, timeout, ttype) : pass
    @ASYNC
    def wait_async                (self, timeout, ttype) : pass

    @SYNC
    def get_state                 (self, ttype)          : pass
    @ASYNC
    def get_state_async           (self, ttype)          : pass

    @SYNC
    def get_result                (self, ttype)          : pass
    @ASYNC
    def get_result_async          (self, ttype)          : pass

    @SYNC
    def get_object                (self, ttype)          : pass
    @ASYNC
    def get_object_async          (self, ttype)          : pass

    @SYNC
    def re_raise                  (self, ttype)          : pass
    @ASYNC
    def re_raise_async            (self, ttype)          : pass


    #-----------------------------------------------------------------
    # attribute getters
    @SYNC
    def get_exit_code             (self, ttype)          : pass
    @ASYNC
    def get_exit_code_async       (self, ttype)          : pass

    @SYNC
    def get_created               (self, ttype)          : pass
    @ASYNC
    def get_created_async         (self, ttype)          : pass

    @SYNC
    def get_started               (self, ttype)          : pass
    @ASYNC
    def get_started_async         (self, ttype)          : pass

    @SYNC
    def get_finished              (self, ttype)          : pass
    @ASYNC
    def get_finished_async        (self, ttype)          : pass

    @SYNC
    def get_execution_hosts       (self, ttype)          : pass
    @ASYNC
    def get_execution_hosts_async (self, ttype)          : pass


# class Self (Job, monitoring.Steerable) :
class Self (Job) :

    @SYNC
    def init_instance             (self)                 : pass
    @ASYNC
    def init_instance_async       (self)                 : pass








__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


""" SAGA job interface """

import radical.utils            as ru
import radical.utils.signatures as rus

from .constants  import *
from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base    as sab

from .. import attributes       as sa
from .. import exceptions       as se
from .. import sasync
from .. import task             as st
from .. import base             as sb

from .  import description      as descr


# ------------------------------------------------------------------------------
#
class Job (sb.Base, st.Task, sasync.Async) :
    '''Represents a SAGA job as defined in GFD.90

    A 'Job' represents a running application instance, which may consist of one
    or more processes.  Jobs are created by submitting a Job description to
    a Job submission system -- usually a queuing system, or some other service
    which spawns jobs on the user's behalf.

    Jobs have a unique ID (see get_job_id()), and are stateful entities -- their
    'state' attribute changes according to a well defined state model:

    A job as returned by job.Service.create(jd) is in 'New' state -- it is not
    yet submitted to the job submission backend.  Once it was submitted, via
    run(), it will enter the 'Pending' state, where it waits to get actually
    executed by the backend (e.g. waiting in a queue etc).  Once the job is
    actually executed, it enters the 'Running' state -- only in that state is
    the job actually consuming resources (CPU, memory, ...).

    Jobs can leave the 'Running' state in three different ways: they finish
    successfully on their own ('Done'), they finish unsuccessfully on their own,
    or get canceled by the job management backend ('Failed'), or they get
    actively canceled by the user or the application ('Canceled').

    The methods defined on the Job object serve two purposes: inspecting the
    job's state, and initiating job state transitions.

    '''

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  rus.optional (str),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, _method_type='run', _adaptor=None, _adaptor_state={}, _ttype=None) :
        '''
        _adaptor`` references the adaptor class instance which created this
        task instance.

        The ``_method_type`` parameter is flattened into the job constructor to
        satisfy the bulk optimization properties of the saga.Task class, whose
        interface is implemented by saga.job.Job. ``_method_type`` specifies
        the SAGA API method which task is representing.  For jobs, that is the
        'run' method.

        We don't have a create classmethod -- jobs are never constructed by the
        user
        '''

        if not _adaptor :
            raise se.IncorrectState ("saga.job.Job constructor is private")

        self._valid = False


        # we need to keep _method_type around, for the task interface (see
        # :class:`saga.Task`)
        self._method_type = _method_type

        # We need to specify a schema for adaptor selection -- and
        # simply choose the first one the adaptor offers.
        schema = _adaptor.get_schemas()[0]
        if  'job_schema' in _adaptor_state :
            schema = _adaptor_state['job_schema']

        self._base = super  (Job, self)
        self._base.__init__ (schema, _adaptor, _adaptor_state, ttype=None)


        # set attribute interface properties
        self._attributes_allow_private (True)
        self._attributes_extensible    (False)
        self._attributes_camelcasing   (True)

        # register properties with the attribute interface
        self._attributes_register   (STATE,            UNKNOWN, sa.ENUM,   sa.SCALAR, sa.READONLY)
        self._attributes_register   (EXIT_CODE,        None,    sa.INT,    sa.SCALAR, sa.READONLY)
        self._attributes_register   (CREATED,          None,    sa.INT,    sa.SCALAR, sa.READONLY)
        self._attributes_register   (STARTED,          None,    sa.INT,    sa.SCALAR, sa.READONLY)
        self._attributes_register   (FINISHED,         None,    sa.INT,    sa.SCALAR, sa.READONLY)
        self._attributes_register   (EXECUTION_HOSTS,  None,    sa.STRING, sa.VECTOR, sa.READONLY)
        self._attributes_register   (ID,               None,    sa.STRING, sa.SCALAR, sa.READONLY)
        self._attributes_register   (NAME,             None,    sa.STRING, sa.SCALAR, sa.READONLY)
        self._attributes_register   (SERVICE_URL,      None,    sa.URL,    sa.SCALAR, sa.READONLY)

        self._attributes_set_enums  (STATE, [UNKNOWN, NEW,     PENDING,  RUNNING,
                                             DONE,    FAILED,  CANCELED, SUSPENDED])

        self._attributes_set_getter (STATE,           self.get_state)
        self._attributes_set_getter (ID,              self.get_id)
        self._attributes_set_getter (NAME,            self.get_name)
        self._attributes_set_getter (EXIT_CODE,       self._get_exit_code)
        self._attributes_set_getter (CREATED,         self._get_created)
        self._attributes_set_getter (STARTED,         self._get_started)
        self._attributes_set_getter (FINISHED,        self._get_finished)
        self._attributes_set_getter (EXECUTION_HOSTS, self._get_execution_hosts)
        self._attributes_set_getter (SERVICE_URL    , self._get_service_url)

        self._valid = True


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job')
    @rus.returns (str)
    def __str__  (self) :
        """
        __str__()

        String representation. Returns the job's ID.
        """

        if  not self._valid :
            return 'no job id'

        return str (self.id)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, str, st.Task))
    def get_id   (self, ttype=None) :
        """
        get_id()

        Return the job ID.
        """
        id = self._adaptor.get_id (ttype=ttype)
        return id


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, str, st.Task))
    def get_name (self, ttype=None) :
        """
        get_name()

        Return the job name.
        """
        name = self._adaptor.get_name(ttype=ttype)
        return name

    # --------------------------------------------------------------------------
    #
    @rus.takes          ('Job',
                         rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns        ((str, st.Task))
    def get_description (self, ttype=None) :
        """
        get_description()

        Return the job description this job was created from.

        The returned description can be used to inspect job properties
        (executable name, arguments, etc.).  It can also be used to start
        identical job instances.

        The returned job description will in general reflect the actual state of
        the running job, and is not necessarily a simple copy of the job
        description which was used to create the job instance.  For example, the
        environment variables in the returned job description may reflect the
        actual environment of the running job instance.


        **Example**::


          service = saga.job.Service("fork://localhost")
          jd = saga.job.Description ()
          jd.executable = '/bin/date'

          j1 = service.create_job(jd)
          j1.run()

          j2 = service.create_job(j1.get_description())
          j2.run()

          service.close()
        """
        return self._adaptor.get_description (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes    ('Job',
                   rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns  ((str, st.Task))
    def get_stdin (self, ttype=None) :
        """
        get_stdin()

        ttype:     saga.task.type enum
        ret:       string / saga.Task

        Return the job's STDIN as string.
        """

        # FIXME: we have no means to set a stdin stream
        return self._adaptor.get_stdin (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((str, st.Task))
    def get_stdout (self, ttype=None) :
        """
        get_stdout()

        ttype:     saga.task.type enum
        ret:       string / saga.Task

        Return the job's STDOUT as string.
        """
        return self._adaptor.get_stdout (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((str, st.Task))
    def get_stdout_string (self, ttype=None) :
        """
        get_stdout_string()

        Return the job's STDOUT as string.

        ttype:     saga.task.type enum
        ret:       string / saga.Task

        THIS METHOD IS DEPRECATED AND WILL BE REMOVED IN A FUTURE RELEASE.
        USE job.get_stdout() INSTEAD.
        """
        return self.get_stdout (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((str, st.Task))
    def get_stderr (self, ttype=None) :
        """
        get_stderr()

        Return the job's STDERR as string.

        ttype:     saga.task.type enum
        ret:       string / saga.Task
        """
        return self._adaptor.get_stderr (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((str, st.Task))
    def get_stderr_string (self, ttype=None) :
        """
        get_stderr_string()

        Return the job's STDERR.

        ttype:     saga.task.type enum
        ret:       string / saga.Task

        THIS METHOD IS DEPRECATED AND WILL BE REMOVED IN A FUTURE RELEASE.
        USE job.get_stderr() INSTEAD.
        """
        return self.get_stderr(ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((str, st.Task))
    def get_log (self, ttype=None) :
        """
        get_log_string()

        Return the job's log information, ie. backend specific log messages
        which have been collected during the job execution.  Those messages also
        include stdout/stderr from the job's pre- and post-exec.  The returned
        string generally contains one log message per line, but the format of
        the string is ultimately undefined.

        ttype:     saga.task.type enum
        ret:       string / saga.Task
        """
        return self._adaptor.get_log (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((str, st.Task))
    def get_log_string (self, ttype=None) :
        """
        get_log_string()

        Return the job's log information, ie. backend specific log messages
        which have been collected during the job execution.  Those messages also
        include stdout/stderr from the job's pre- and post-exec.  The returned
        string generally contains one log message per line, but the format of
        the string is ultimately undefined.

        ttype:     saga.task.type enum
        ret:       string / saga.Task

        THIS METHOD IS DEPRECATED AND WILL BE REMOVED IN A FUTURE RELEASE.
        USE job.get_log() INSTEAD.
        """
        return self.get_log (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def suspend  (self, ttype=None) :
        """
        suspend()

        Suspend the job.
        """
        return self._adaptor.suspend (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def resume   (self, ttype=None) :
        """
        resume()

        Resume the job.
        """
        return self._adaptor.resume (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((rus.nothing, st.Task))
    def checkpoint (self, ttype=None) :
        """
        checkpoint()

        Checkpoint the job.
        """
        return self._adaptor.checkpoint (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  descr.Description,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def migrate  (self, jd, ttype=None) :
        """
        jd:        saga.job.Description
        ttype:     saga.task.type enum
        ret:       None / saga.Task
        """
        return self._adaptor.migrate (jd, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  int,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def signal   (self, signum, ttype=None) :
        """
        signal(signum)

        Send a signal to the job.

        :param signum: signal to send
        :type  signum: int
        """
        return self._adaptor.signal (signum, ttype=ttype)


    #-----------------------------------------------------------------
    #
    id          = property (get_id)            # string
    description = property (get_description)   # Description
   #stdin       = property (get_stdin)         # File
    stdout      = property (get_stdout_string) # string
    stderr      = property (get_stderr_string) # string
    log         = property (get_log)           # string


    #-----------------------------------------------------------------
    #
    # task methods flattened into job :-/
    #
    @rus.takes   ('Job',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def run      (self, ttype=None) :
        """
        run()

        Run (start) the job.

        Request that the job is being executed by the backend.  If the backend
        is accepting this run request, the job will move to the 'Pending' or
        'Running' state -- otherwise this method will raise an error, and the
        job will be moved to 'Failed'.


        **Example**::

            js = saga.job.Service("fork://localhost")
            jd = saga.job.Description ()
            jd.executable = '/bin/date'
            j  = js.create_job(jd)

            if j.get_state() == saga.job.NEW :
                print("new")
            else :
                print("oops!")

            j.run()

            if   j.get_state() == saga.job.PENDING :
                print("pending")
            elif j.get_state() == saga.job.RUNNING :
                print("running")
            else :
                print("oops!")
          """

        return self._adaptor.run (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  float,
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def cancel   (self, timeout=None, ttype=None) :
        """
        cancel(timeout)

        Cancel the execution of the job.

        :param timeout: `cancel` will return after timeout
        :type  timeout: float

        **Example**::

          js = saga.job.Service("fork://localhost")
          jd = saga.job.Description ()
          jd.executable = '/bin/date'
          j  = js.create_job(jd)

          if   j.get_state() == saga.job.NEW :
              print("new")
          else :
              print("oops!")

          j.run()

          if   j.get_state() == saga.job.PENDING :
              print("pending")
          elif j.get_state() == saga.job.RUNNING :
              print("running")
          else :
              print("oops!")

          j.cancel()

          if   j.get_state() == saga.job.CANCELED :
              print("canceled")
          else :
              print("oops!")
        """
        return self._adaptor.cancel (timeout, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Job',
                  rus.optional (float),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def wait     (self, timeout=None, ttype=None) :
        """
        wait(timeout)

        :param timeout: `wait` will return after timeout
        :type  timeout: float

        Wait for a running job to finish execution.

        The optional timeout parameter specifies the time to wait, and accepts
        the following values::

          timeout <  0  : wait forever (block) -- same for 'None'
          timeout == 0  : wait not at all (non-blocking test)
          timeout >  0  : wait for 'timeout' seconds

        On a non-negative timeout, the call can thus return even if the job is
        not in final state, and the application should check the actual job
        state.  The default timeout value is 'None' (blocking).


        **Example**::

          js = saga.job.Service("fork://localhost")
          jd = saga.job.Description ()
          jd.executable = '/bin/date'
          j  = js.create_job(jd)

          if   j.get_state() == saga.job.NEW :
              print("new")
          else :
              print("oops!")

          j.run()

          if   j.get_state() == saga.job.PENDING :
              print("pending")
          elif j.get_state() == saga.job.RUNNING :
              print("running")
          else :
              print("oops!")

          j.wait(-1.0)

          if   j.get_state() == saga.job.DONE :
              print("done")
          elif j.get_state() == saga.job.FAILED :
              print("failed")
          else :
              print("oops!")
        """

        if  None == timeout :
            timeout = -1.0 # FIXME

        return self._adaptor.wait (timeout, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes    ('Job',
                   rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns  ((rus.one_of (UNKNOWN, NEW, PENDING, RUNNING, SUSPENDED, DONE, FAILED, CANCELED), st.Task))
    def get_state (self, ttype=None) :
        """
        get_state()

        Return the current state of the job.

        **Example**::

          js = saga.job.Service("fork://localhost")
          jd = saga.job.Description ()
          jd.executable = '/bin/date'
          j  = js.create_job(jd)

          if   j.get_state() == saga.job.NEW :
              print("new")
          else :
              print("oops!")

          j.run()

          if   j.get_state() == saga.job.PENDING :
              print("pending")
          elif j.get_state() == saga.job.RUNNING :
              print("running")
          else :
              print("oops!")
        """
        return self._adaptor.get_state (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((rus.anything, st.Task))
    def get_result (self, ttype=None) :
        """
        get_result()
        """
        return self._adaptor.get_result (ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((sb.Base, st.Task))
    def get_object (self, ttype=None) :
        """ :todo: describe me
            :note: this will return the job_service which created the job.
        """
        return self._adaptor.get_object (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job',
                    rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns   ((se.SagaException, st.Task))
    def get_exception (self, ttype=None) :
        """ :todo: describe me

            :note: if job failed, that will get an exception describing
                   why, if that exists.  Otherwise, the call returns None.
        """

        if  self.state == FAILED :
            return se.NoSuccess ("job stderr: %s" % self.get_stderr_string ())
        else :
            return None


    # --------------------------------------------------------------------------
    #
    @rus.takes     ('Job')
    @rus.returns   (rus.nothing)
    def re_raise   (self) :
        """ :todo: describe me

            :note: if job failed, that will re-raise an exception describing
                   why, if that exists.  Otherwise, the call does nothing.
        """

        if  self.state == FAILED :
            raise se.NoSuccess ("job stderr: %s" % self.get_stderr_string ())
        else :
            return


    # ----------------------------------------------------------------
    #
    # attribute getters
    #
    @rus.takes         ('Job',
                        rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns       ((rus.nothing, int, st.Task))
    def _get_exit_code (self, ttype=None) :
        ec = self._adaptor.get_exit_code(ttype=ttype)
        if ec in [None, ""]:
            return None
        else:
            return ec

    # --------------------------------------------------------------------------
    #
    @rus.takes       ('Job',
                      rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns     ((rus.nothing, float, st.Task))
    def _get_created (self, ttype=None) :
        return self._adaptor.get_created (ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes       ('Job',
                      rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns     ((rus.nothing, float, st.Task))
    def _get_started (self, ttype=None) :
        return self._adaptor.get_started (ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes       ('Job',
                      rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns     ((rus.nothing, float, st.Task))
    def _get_finished (self, ttype=None) :
        return self._adaptor.get_finished (ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes       ('Job',
                      rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns     ((rus.nothing, rus.list_of (str), st.Task))
    def _get_execution_hosts (self, ttype=None) :
        return self._adaptor.get_execution_hosts (ttype=ttype)

    # --------------------------------------------------------------------------
    #
    @rus.takes       ('Job',
                      rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns     ((rus.nothing, ru.Url, st.Task))
    def _get_service_url (self, ttype=None) :
        return self._adaptor.get_service_url (ttype=ttype)

    name      = property (get_name)        # job name
    state     = property (get_state)       # state enum
    result    = property (get_result)      # result type    (None)
    object    = property (get_object)      # object type    (job_service)
    exception = property (re_raise)        # exception type


    # ----------------------------------------------------------------
    #
    ExitCode = property (doc = """
    The job's exitcode.

    this attribute is only meaningful if the job is in 'Done' or 'Final'
    state - for all other job states, this attribute value is undefined.

    **Example**::


      js = saga.job.Service("fork://localhost")
      jd = saga.job.Description ()
      jd.executable = '/bin/date'
      j  = js.create_job(jd)

      j.run()
      j.wait()

      if j.get_state() == saga.job.FAILED :
        if j.exitcode == "42" :
            print("Ah, galaxy bypass error!")
        else :
            print("oops!")

    """)


    # ----------------------------------------------------------------
    #
    JobID = property (doc = """
    The job's identifier.

    This attribute is equivalent to the value returned by job.get_job_id()
    """)


    # ----------------------------------------------------------------
    #
    ServiceURL = property (doc = """
    The URL of the :class:`saga.job.Service` instance managing this job.

    This attribute is represents the URL under where the job management
    service can be contacted which owns the job.  The value is equivalent to
    the service part of the job_id.

    **Example**::


      js = saga.job.Service("fork://localhost")
      jd = saga.job.Description ()
      jd.executable = '/bin/date'
      j  = js.create_job(jd)

      if j.serviceurl == "fork://localhost" :
          print("yes!")
      else :
          print("oops!")

    """)


# ------------------------------------------------------------------------------
#
# class Self (Job, monitoring.Steerable) :
#
class Self (Job) :

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Self',
                  rus.optional (str),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, _method_type='run', _adaptor=None, _adaptor_state={}, _ttype=None) :


        self._base = super  (Job, self)
        self._base.__init__ (_method_type, _adaptor, _adaptor_state, _ttype=_ttype)








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils            as ru
import radical.utils.signatures as rus

from ..constants import SYNC, ASYNC, TASK
from ..adaptors  import base    as sab

from .. import session          as ss
from .. import task             as st
from .. import base             as sb
from .. import sasync


# ------------------------------------------------------------------------------
#
class Entry (sb.Base, sasync.Async) :
    '''
    Represents a SAGA namespace entry as defined in GFD.90

    The saga.namespace.Entry class represents, as the name indicates,
    an entry in some (local or remote) namespace.  That class offers
    a number of operations on that entry, such as copy, move and remove::

        # get an entry handle
        entry = saga.namespace.Entry ("sftp://localhost/tmp/data/data.bin")

        # copy the entry
        entry.copy ("sftp://localhost/tmp/data/data.bak")

        # move the entry
        entry.move ("sftp://localhost/tmp/data/data.new")
    '''

    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (sab.Base),
                  rus.optional (dict),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (rus.nothing)
    def __init__ (self, url=None, flags=None, session=None,
                  _adaptor=None, _adaptor_state={}, _ttype=None) :
        '''
        :param url: Url of the (remote) entry
        :type  url: :class:`saga.Url`

        flags:     flags enum
        session:   saga.Session
        ret:       obj

        Construct a new entry object

        The specified entry is expected to exist -- otherwise a DoesNotExist
        exception is raised.  Also, the URL must point to an entry (not to
        a directory), otherwise a BadParameter exception is raised.

        Example::

            # get an entry handle
            entry = saga.namespace.Entry("sftp://localhost/tmp/data/data.bin")

            # print the entry's url
            print(entry.get_url ())
        '''

        self._session      = session
        self._is_recursive = False  # recursion guard (FIXME: NOT THREAD SAFE)

        # param checks
        if not session :
            session = ss.Session (default=True)

        if  not flags : flags = 0
        url     = ru.Url (url)
        scheme  = url.scheme.lower ()

        self._base = super  (Entry, self)
        self._base.__init__ (scheme, _adaptor, _adaptor_state,
                             url, flags, session, ttype=_ttype)


    # --------------------------------------------------------------------------
    #
    @classmethod
    @rus.takes   ('Entry',
                  rus.optional ((ru.Url, str)),
                  rus.optional (int, rus.nothing),
                  rus.optional (ss.Session),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns (st.Task)
    def create   (cls, url=None, flags=None, session=None, ttype=None) :
        '''
        url:       saga.Url
        flags:     saga.namespace.flags enum
        session:   saga.Session
        ttype:     saga.task.type enum
        ret:       saga.Task
        '''

        # param checks
        if  not flags : flags = 0
        if not session :
            session = ss.Session (default=True)

        return cls (url, flags, session, _ttype=ttype)._init_task



    # ----------------------------------------------------------------
    #
    # namespace entry methods
    #
    @rus.takes   ('Entry',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((ru.Url, st.Task))
    def get_url  (self, ttype=None) :
        '''
        ttype:         saga.task.type enum
        ret:           saga.Url / saga.Task

        Return the complete url pointing to the entry.

        The call will return the complete url pointing to
        this entry as a saga.Url object::

            # print URL of an entry
            entry = saga.namespace.Entry("sftp://localhost/etc/passwd")
            print(entry.get_url())
        '''
        return self._adaptor.get_url (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((str, st.Task))
    def get_cwd  (self, ttype=None) :
        '''
        ttype:         saga.task.type enum
        ret:           string / saga.Task
        '''
        return self._adaptor.get_cwd (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((str, st.Task))
    def get_name (self, ttype=None) :
        '''
        ttype:         saga.task.type enum
        ret:           string / saga.Task
        '''
        return self._adaptor.get_name (ttype=ttype)


    # ----------------------------------------------------------------
    #
    # namespace entry / directory methods
    #
    @rus.takes   ('Entry',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_dir   (self, ttype=None) :
        '''
        ttype:         saga.task.type enum
        ret:           bool / saga.Task

        Returns True if path is a directory, False otherwise.

        Example::

            # inspect an entry
            dir  = saga.namespace.Directory("sftp://localhost/tmp/")
            if dir.is_dir ('data'):
                # do something
        '''
        return self._adaptor.is_dir_self (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_entry (self, ttype=None) :
        '''
        ttype:         saga.task.type enum
        ret:           bool / saga.Task
        '''
        return self._adaptor.is_entry_self (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((bool, st.Task))
    def is_link  (self, ttype=None) :
        '''
        tgt:           saga.Url / None
        ttype:         saga.task.type enum
        ret:           bool / saga.Task
        '''
        return self._adaptor.is_link_self (ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes    ('Entry',
                   rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns  ((ru.Url, st.Task))
    def read_link (self, ttype=None) :
        '''
        tgt:           saga.Url / None
        ttype:         saga.task.type enum
        ret:           saga.Url / saga.Task
        '''

        return self._adaptor.read_link_self (ttype=ttype)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def copy     (self, tgt, flags=0, ttype=None) :
        '''
        tgt:           saga.Url
        flags:         enum flags
        ttype:         saga.task.type enum
        ret:           None / saga.Task

        Copy the entry to another location

        :param target: Url of the copy target.
        :param flags: Flags to use for the operation.

        The entry is copied to the given target location.  The target URL must
        be an absolute path, and can be a target entry name or target
        directory name.  If the target entry exists, it is overwritten::

            # copy an entry
            entry = saga.namespace.Entry("sftp://localhost/tmp/data/data.bin")
            entry.copy ("sftp://localhost/tmp/data/data.bak")
        '''

        # parameter checks
        if  not flags : flags = 0
        return self._adaptor.copy_self (tgt, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def link     (self, tgt, flags=0, ttype=None) :
        '''
        tgt:           saga.Url
        flags:         enum flags
        ttype:         saga.task.type enum
        ret:           None / saga.Task
        '''

        if  not flags : flags = 0
        return self._adaptor.link_self (tgt, flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  (ru.Url, str),
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def move     (self, tgt, flags=0, ttype=None) :
        '''
        :param target: Url of the move target.
        :param flags:  Flags to use for the operation.

        ttype:         saga.task.type enum
        ret:           None / saga.Task

        Move the entry to another location

        The entry is copied to the given target location.  The target URL must
        be an absolute path, and can be a target entry name or target
        directory name.  If the target entry exists, it is overwritten::

            # copy an entry
            entry = rs.namespace.Directory("sftp://localhost/tmp/data/data.bin")
            entry.move ("sftp://localhost/tmp/data/data.bak")
        '''
        if  not flags : flags = 0
        return self._adaptor.move_self (tgt, flags, ttype=ttype)



    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional (int, rus.nothing),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def remove   (self, flags=0, ttype=None) :
        '''
        :param flags:  Flags to use for the operation.

        ttype:         saga.task.type enum
        ret:           None / saga.Task

        Reove the entry.

        The entry is removed, and this object instance is then invalid for
        further operations.

            # remove an entry
            entry = rs.namespace.Directory("sftp://localhost/tmp/data/data.bin")
            entry.remove ()
        '''
        if  not flags : flags = 0
        return self._adaptor.remove_self (flags, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Entry',
                  rus.optional (float),
                  rus.optional (rus.one_of (SYNC, ASYNC, TASK)))
    @rus.returns ((rus.nothing, st.Task))
    def close (self, timeout=None, ttype=None) :
        '''
        timeout:       float
        ttype:         saga.task.type enum
        ret:           None / saga.Task
        '''
        return self._adaptor.close (timeout, ttype=ttype)


    # --------------------------------------------------------------------------
    #
    url  = property (get_url)   # saga.Url
    cwd  = property (get_cwd)   # string
    name = property (get_name)  # string


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import time
import threading as mt
from collections import OrderedDict

CACHE_DEFAULT_SIZE = 10000
CACHE_DEFAULT_TTL  = 1.0    # 1 second

VAL = 'val'
TTL = 'ttl'

######################################################################
#
class Cache :

    # ----------------------------------------------------------------
    #
    def __init__ (self, logger, size=CACHE_DEFAULT_SIZE, ttl=CACHE_DEFAULT_TTL) :

        if int (size) < 1 :
            raise AttributeError ('size < 1 or not a number')

        if int (ttl) < 0 :
            raise AttributeError ('ttl < 0 or not a number')

        self.size   = size
        self.ttl    = ttl
        self.dict   = OrderedDict()
        self.lock   = mt.RLock()
        self.logger = logger
        self.hit    = 0
        self.miss   = 0

        # start a thread which, with low priority, cleans out the dict now and
        # then (pops items until a live one is found


    # ----------------------------------------------------------------
    #
    def _dump (self) :
        print(" ---------------------------------------------- ")
        print(" CACHE STATISTICS : ")
        print(" size: %5d" % len(self.dict))
        print(" hit : %5d" % self.hit) 
        print(" miss: %5d" % self.miss)
        print(list(self.dict.keys()))
        print(" ---------------------------------------------- ")


    # ----------------------------------------------------------------
    #
    def get (self, key) :

        self.logger.debug ("redis_cache_get %s", key)

        with self.lock:

            # check if we have a live entry
            if key in self.dict :

                now = time.time ()

                if self.ttl and self.dict[key][TTL] > now :
                    # if yes, cache hit!
                    # return data -- doh!
                    self.hit += 1
                    return self.dict[key][VAL]

                else :
                    # entry timed out
                    del self.dict[key]

            # cache entry not found, or timed out
            self.miss += 1
            raise AttributeError ("cache miss for '%s' " % key)


    # ----------------------------------------------------------------
    #
    def set (self, key, value) :

        with self.lock :

            # remove superfluous(?) entries
            while len (self.dict) >= self.size :
                self.dict.popitem (last=False)

            self.dict[key]      = {}
            self.dict[key][VAL] = value
            self.dict[key][TTL] = time.time () + self.ttl


    # ----------------------------------------------------------------
    #
    def delete (self, key) :

        with self.lock :
            del self.dict[key]


    # ----------------------------------------------------------------








__author__    = "Andre Merzky"
__copyright__ = "Copyright 2020, The SAGA Project"
__license__   = "MIT"


''' no operation (noop) based job adaptor implementation '''

import time
import threading

from ...               import exceptions as rse
from ..                import base
from ..cpi             import SYNC_CALL
from ..cpi             import job as cpi
from ...               import job as api


# ------------------------------------------------------------------------------
#
class _job_state_monitor(threading.Thread):

    # --------------------------------------------------------------------------
    #
    def __init__(self, log):

        self._log  = log
        self._lock = threading.Lock()
        self._term = threading.Event()
        self._jobs = dict()
        self._cnt  = 0

        super(_job_state_monitor, self).__init__()

        self.setDaemon(True)


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._term.set()


    # --------------------------------------------------------------------------
    #
    def add_job(self, job):

        job._id = 'job.%06d' % self._cnt
        self._cnt += 1

        assert(job._id not in self._jobs)
        job._set_state(api.RUNNING)
        job._started = time.time()

        with self._lock:
            self._jobs[job._id] = job


    # --------------------------------------------------------------------------
    #
    def get_job(self, jid):

        assert(jid in self._jobs)
        with self._lock:
            return self._jobs[jid]


    # --------------------------------------------------------------------------
    #
    def list_jobs(self):

        with self._lock:
            return list(self._jobs.keys())


    # --------------------------------------------------------------------------
    #
    def run(self):

        try:

            while not self._term.is_set():

                now  = time.time()
                keep = dict()

                with self._lock:

                    for job in self._jobs.values():

                        if job.get_state() == api.CANCELED:
                            continue

                        if job.tgt < now:
                            job._finished  = now
                            job._exit_code = 0
                            job._set_state(api.DONE)

                        else:
                            keep[job._id] = job

                    self._jobs = keep

                time.sleep(0.1)

        except Exception:

            self._log.exception("Exception in job monitoring thread")


# ------------------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.noop_job"
_ADAPTOR_SCHEMAS       = ["noop"]



# ------------------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES  = {
    "jdes_attributes" : [api.NAME,
                         api.EXECUTABLE,
                         api.PRE_EXEC,
                         api.POST_EXEC,
                         api.ARGUMENTS,
                         api.ENVIRONMENT,
                         api.WORKING_DIRECTORY,
                         api.FILE_TRANSFER,
                         api.INPUT,
                         api.OUTPUT,
                         api.ERROR,
                         api.NAME,
                         api.WALL_TIME_LIMIT,
                         api.TOTAL_CPU_COUNT,
                         api.TOTAL_GPU_COUNT,
                         api.PROCESSES_PER_HOST,
                         api.SPMD_VARIATION,
                        ],
    "job_attributes"  : [api.EXIT_CODE,
                         api.EXECUTION_HOSTS,
                         api.CREATED,
                         api.STARTED,
                         api.FINISHED],
    "metrics"         : [api.STATE,
                         api.STATE_DETAIL],
    "contexts"        : {}
}

# ------------------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC           = {
    "name"            : _ADAPTOR_NAME,
    "capabilities"    : _ADAPTOR_CAPABILITIES,
    "description"     : '''
        The Noop job adaptor, which fakes job execution and is used for testing
        and benchmarking purposes.''',
    "example"         : "examples/jobs/noopjob.py",
    "schemas"         : {"noop"  : "fake job execution"}
}

# ------------------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA

_ADAPTOR_INFO = {
    "name"            : _ADAPTOR_NAME,
    "version"         : "v0.1",
    "schemas"         : _ADAPTOR_SCHEMAS,
    "capabilities"    : _ADAPTOR_CAPABILITIES,
    "cpis"            : [{
                                 "type" : "radical.saga.job.Service",
                                 "class": "NoopJobService"
                         },{
                                 "type" : "radical.saga.job.Job",
                                 "class": "NoopJob"
                         }]
}


# ------------------------------------------------------------------------------
# The adaptor class
class Adaptor(base.Base):
    '''
    This is the actual adaptor class, which gets loaded by SAGA (i.e., by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    '''


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        base.Base.__init__(self, _ADAPTOR_INFO, expand_env=False)


    # --------------------------------------------------------------------------
    #
    def sanity_check(self):

        pass


# ------------------------------------------------------------------------------
#
class NoopJobService(cpi.Service):

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super (NoopJobService, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    def __del__(self):

        try   : self.close()
        except: pass


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        '''
        Service instance constructor
        '''

        self.rm      = rm_url
        self.session = session
        self.jobs    = dict()

        # Use `_set_session` method of the base class to set the session object.
        # `_set_session` and `get_session` methods are provided by `CPIBase`.
        self._set_session(session)

        # the monitoring thread - one per service instance.
        self.monitor = _job_state_monitor(self._logger)
        self.monitor.start()

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    def close(self):

        if self.monitor:
            self.monitor.stop()


    # --------------------------------------------------------------------------
    #
    #
    def _job_run(self, job):
        '''
        runs a job on the wrapper via pty, and returns the job id
        '''

        if not job.jd.executable.endswith('sleep'):
            raise ValueError('expected "sleep", not %s' % job.jd.executable)


        if len(job.jd.arguments) != 1:
            raise ValueError('expected int argument, not %s' % job.jd.arguments)

        job.tgt = time.time() + int(job.jd.arguments[0])
        self.monitor.add_job(job)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run_job(self, cmd, host):
        '''
        Implements adaptors.cpi.job.Service.run_job()
        '''

        if not cmd:
            raise rse.BadParameter._log(self._logger,
                    "run_job needs a command to run")

        if host and host != self.rm.host:
            raise rse.BadParameter._log(self._logger,
                    "Can only run jobs on %s, not on %s" % (self.rm.host, host))

        jd = api.Description()

        exe, arg = cmd.split()

        jd.executable = exe
        jd.arguments  = [arg]

        job = self.create_job(jd)
        job.run()

        return job


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service"    : self,
                         "job_description": jd,
                         "job_schema"     : self.rm.schema }

        return api.Job(_adaptor=self._adaptor, _adaptor_state=adaptor_state)


    # --------------------------------------------------------------------------
    @SYNC_CALL
    def get_url(self):

        return self.rm


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):

        return self.monitor.list_jobs()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, job_id, no_reconnect=False):

        return self.monitor.get_job(job_id)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_run(self, jobs):
        '''
        From all the job descriptions in the container, build a bulk, and submit
        as async.
        '''

        self._logger.debug("container run: %s"  %  str(jobs))

        for job in jobs:
            job.run()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_wait(self, jobs, mode, timeout):

        for job in jobs:

            job.wait()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_cancel(self, jobs, timeout):

        for job in jobs:

            job.cancel()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def container_get_states(self, jobs):

        return [job.get_state() for job in jobs]


# ------------------------------------------------------------------------------
#
class NoopJob(cpi.Job):

    # --------------------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        _cpi_base = super (NoopJob, self)
        _cpi_base.__init__(api, adaptor)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, job_info):

        if 'job_description' in job_info:
            # comes from job.service.create_job()
            self.js = job_info["job_service"]
            self.jd = job_info["job_description"]

            # initialize job attribute values
            self._id              = None
            self._name            = self.jd.get(api.NAME)
            self._log             = list()
            self._state           = None
            self._exit_code       = None
            self._exception       = None
            self._created         = time.time()
            self._name            = self.jd.name
            self._started         = None
            self._finished        = None

            self._set_state(api.NEW)

        elif 'job_id' in job_info:
            # initialize job attribute values
            self.js               = job_info["job_service"]
            self.jd               = None
            self._id              = job_info['job_id']
            self._name            = job_info.get('job_name')
            self._log             = list()
            self._state           = None
            self._exit_code       = None
            self._exception       = None
            self._created         = None
            self._name            = None
            self._started         = None
            self._finished        = None

        else:
            # don't know what to do...
            raise rse.BadParameter("insufficient info for job creation")

        if self._created: self._created = float(self._created)

        return self.get_api()


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description(self):

        return self.jd


    # --------------------------------------------------------------------------
    #
    def _update_state(self, state):

        # update state, and report to application
        self._state = state
        self._api()._attributes_i_set('state', self._state, self._api()._UP)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        '''
        Implements adaptors.cpi.job.Job.get_state()
        '''

        # may not yet have backend representation, state is 'NEW'
        if self._id is None:
            return self._state

        return self._state


    # --------------------------------------------------------------------------
    #
    def _set_state(self, state):

        old_state = self._state

        # on state changes, trigger notifications
        if old_state != state:
            self._state  = state
            self._api()._attributes_i_set('state', state, self._api()._UP)

        return self._state


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):

        return self._created


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):

        return self._started


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):

        return self._finished


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_stdout(self):

        return ''


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_stderr(self):

        return ''


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_log(self):

        return ''


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):

        if not self.js:
            raise rse.IncorrectState("Job Service URL unknown")
        else:
            return self.js.get_url()


    # --------------------------------------------------------------------------
    #
    # TODO: this should also fetch the(final) state, to safe a hop
    # TODO: implement via notifications
    #
    @SYNC_CALL
    def wait(self, timeout):
        '''
        A call to the noop to do the WAIT would block the noop for any
        other interactions.  In particular, it would practically kill it if the
        Wait waits forever...

        So we implement the wait via a state pull.  The *real* solution is to
        implement state notifications, and wait for such a notification to
        arrive within timeout seconds...
        '''

        time_start = time.time()
        time_now   = time_start

        while True:

            state = self.get_state()

            if state in [api.DONE, api.FAILED, api.CANCELED]:
                return True

            # avoid busy poll
            # FIXME: self-tune by checking call latency
            time.sleep(0.1)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):

        return self._id


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name(self):

        return self._name


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):

        return self._exit_code


    # --------------------------------------------------------------------------
    #
    # TODO: the values below should be fetched with every get_state...
    #
    @SYNC_CALL
    def get_execution_hosts(self):

        return [self.js.get_url().host]


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):

        self.js._job_run(self)
        self.js.jobs[self._id] = self._api()

        self._set_state(api.RUNNING)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def suspend(self):

        if self.get_state() != api.RUNNING:
            raise rse.IncorrectState("Cannot suspend, job is not RUNNING")

        self._old_state = self.get_state()
        self._adaptor._update_state(api.SUSPENDED)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def resume(self):

        if self.get_state() != api.SUSPENDED:
            raise rse.IncorrectState("Cannot resume, job is not SUSPENDED")

        self._adaptor._update_state(self._old_state)


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):

        if self.get_state() not in [api.RUNNING,  api.SUSPENDED,
                                    api.CANCELED, api.DONE,
                                    api.FAILED]:
            raise rse.IncorrectState("Cannot cancel, job is not running")

        if self._state in [api.CANCELED, api.DONE, api.FAILED]:
            self._set_state(api.CANCELED)
            return


    # --------------------------------------------------------------------------
    #
    @SYNC_CALL
    def re_raise(self):
        # nothing to do here actually, as run() is synchronous...
        return self._exception


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ashley Z, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


#!/usr/bin/env python2.4
# -*- coding: utf-8 -*-


'''This example shows how to use the iRODS adaptor in 
   some basic ways.  If it executes successfully, the
   iRODS adaptor should work OK on your setup.

   If something doesn't work as expected, try to set 
   SAGA_VERBOSE=3 in your environment before you run the
   script in order to get some debug output.

   If you think you have encountered a defect, please 
   report it at: https://github.com/radical-cybertools/radical.saga/issues
'''

__author__    = "Ashley Zebrowski"
__copyright__ = "Copyright 2012-2013, Ashley Zebrowski"
__license__   = "MIT"

import sys
import os
import logging
import subprocess

import radical.saga as rs


FILE_SIZE     = 1                   # in megs, approx
NUM_REPLICAS  = 1                   # num replicas to create
TEMP_FILENAME = "test.txt"          # filename to create and use for testing
TEMP_DIR      = "/irods_test_dir/"  # directory to create and use for testing


def usage():
    print('Usage: python %s ' % __file__)
    print(' <IRODS_DIRECTORY> (e.x. /osg/home/username/)')
    print(' <IRODS_RESOURCE>  (e.x. osgGridFtpGroup)>')
    print()
    print("For example: python %s " % __file__ + "/irods/mydir/ irodsGroup")
    print("Please specify an iRODS resource group for <IRODS_RESOURCE> " \
          "if possible in order to properly test replication.")


def main(args):

    if len(args) != 3:
        usage()
        exit(-1)

    # directory to store our iRODS files in, don't forget trailing and leading /
    IRODS_DIRECTORY = args[1]

    # iRODS resource or resource group to upload files to
    IRODS_RESOURCE  = args[2]

    # remove any intermediary files that may have been created on iRODS 
    # from an earlier, failed run of this script
    #
    try:
        subprocess.check_call(["irm", IRODS_DIRECTORY + TEMP_FILENAME])
    except:
        pass

    try:
        subprocess.check_call(["irm", '-r', IRODS_DIRECTORY + TEMP_DIR])
    except:
        pass

    try:
        subprocess.check_call(["rm", "/tmp/" + TEMP_FILENAME])
    except:
        pass

    try:
        myfile = rs.replica.LogicalFile('irods://' + IRODS_DIRECTORY +
                                                                  TEMP_FILENAME)

        # grab our home directory (tested on Linux)
        home_dir = os.path.expanduser("~" + "/")
        print("Creating temporary file of size %dM : %s" % \
            (FILE_SIZE, home_dir + TEMP_FILENAME))

        # create a file for us to use with iRODS
        with open(home_dir+TEMP_FILENAME, "wb") as f:
            f.write ("x" * (FILE_SIZE * pow(2,20)) )

        print("Creating iRODS directory object")
        mydir = rs.replica.LogicalDirectory("irods://localhost/" + IRODS_DIRECTORY) 

        print("Uploading file to iRODS")
        myfile = rs.replica.LogicalFile('irods://localhost/'+IRODS_DIRECTORY+TEMP_FILENAME)
        myfile.upload(home_dir + TEMP_FILENAME, \
                     "irods:///this/path/is/ignored/?resource="+IRODS_RESOURCE)

        print("Deleting file locally : %s" % (home_dir + TEMP_FILENAME))
        os.remove(home_dir + TEMP_FILENAME)

        print("Printing iRODS directory listing for %s " % ("irods://localhost/" + IRODS_DIRECTORY))
        for entry in mydir.list():
            print(entry)

        print("Creating iRODS file object")
        myfile = rs.replica.LogicalFile('irods://localhost/' + IRODS_DIRECTORY+TEMP_FILENAME)
        
        print("Creating",NUM_REPLICAS,"replicas for",IRODS_DIRECTORY+TEMP_FILENAME)
        for i in range(NUM_REPLICAS):
            myfile.replicate("irods://localhost/this/path/is/ignored/?resource="+IRODS_RESOURCE)

        print("Locations the file is stored at on iRODS:")
        for entry in myfile.list_locations():
            print(entry)

        print("Downloading logical file %s to /tmp/" % \
            (IRODS_DIRECTORY + TEMP_FILENAME)) 
        myfile.download("/tmp/")

        print("Deleting downloaded file locally: %s" % ("/tmp" + TEMP_FILENAME))
        os.remove("/tmp/" + TEMP_FILENAME)

        print("Making test dir %s on iRODS" % (IRODS_DIRECTORY+TEMP_DIR))
        mydir.make_dir("irods://localhost/"+IRODS_DIRECTORY+TEMP_DIR)
        
        print("Deleting test dir %s from iRODS" % (IRODS_DIRECTORY+TEMP_DIR))
        mydir.remove("irods://localhost/"+IRODS_DIRECTORY+TEMP_DIR)

        print("Deleting file %s from iRODS" % (IRODS_DIRECTORY+TEMP_FILENAME))
        myfile.remove()

    except Exception as ex:
        logging.exception("An error occured while executing the test script!"
                          "Please run with SAGA_VERBOSE=4 set in the"
                          "environment for debug output.\n  %s"
                          % (str(ex)))
        sys.exit(-1)

    print("iRODS test script finished execution")

if __name__ == "__main__":
    sys.exit(main(sys.argv))




__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .. import decorators as cpi_dec
from .. import base       as cpi_base
from .. import sasync     as cpi_async

from .  import entry


SYNC  = cpi_dec.CPI_SYNC_CALL
ASYNC = cpi_dec.CPI_ASYNC_CALL


# ------------------------------------------------------------------------------
# keep order of inheritance!  super() below uses MRO
class Directory (entry.Entry) :

    # ----------------------------------------------------------------
    #
    # initialization methods
    #
    def __init__ (self, api, adaptor) :

        self._cpi_nsentry = super  (Directory, self)
        self._cpi_nsentry.__init__ (api, adaptor)

    @SYNC
    def init_instance           (self, url, flags, session)    : pass
    @ASYNC
    def init_instance_async     (self, url, flags, session)    : pass


    # ----------------------------------------------------------------
    #
    # namespace directory methods
    #
    @SYNC
    def change_dir              (self, url, flags, ttype)      : pass
    @ASYNC
    def change_dir_async        (self, url, flags, ttype)      : pass

    @SYNC
    def list                    (self, npat, ttype)            : pass
    @ASYNC
    def list_async              (self, npat, ttype)            : pass

    @SYNC
    def find                    (self, npat, flags, ttype)     : pass
    @ASYNC
    def find_async              (self, npat, flags, ttype)     : pass

    @SYNC
    def exists                  (self, name, ttype)            : pass
    @ASYNC
    def exists_async            (self, name, ttype)            : pass


    @SYNC
    def is_dir                  (self, name, ttype)            : pass
    @ASYNC
    def is_dir_async            (self, name, ttype)            : pass

    @SYNC
    def is_entry                (self, name, ttype)            : pass
    @ASYNC
    def is_entry_async          (self, name, ttype)            : pass

    @SYNC
    def is_link                 (self, name, ttype)            : pass
    @ASYNC
    def is_link_async           (self, name, ttype)            : pass

    @SYNC
    def read_link               (self, name, ttype)            : pass
    @ASYNC
    def read_link_async         (self, name, ttype)            : pass

    @SYNC
    def get_num_entries         (self, ttype)                  : pass
    @ASYNC
    def get_num_entries_async   (self, ttype)                  : pass

    @SYNC
    def get_entry               (self, num, ttype)             : pass
    @ASYNC
    def get_entry_async         (self, num, ttype)             : pass

    @SYNC
    def copy                    (self, src, tgt, flags, ttype) : pass
    @ASYNC
    def copy_async              (self, src, tgt, flags, ttype) : pass

    @SYNC
    def link                    (self, src, tgt, flags, ttype) : pass
    @ASYNC
    def link_async              (self, src, tgt, flags, ttype) : pass

    @SYNC
    def move                    (self, src, tgt, flags, ttype) : pass
    @ASYNC
    def move_async              (self, src, tgt, flags, ttype) : pass

    @SYNC
    def remove                  (self, tgt, flags, ttype)      : pass
    @ASYNC
    def remove_async            (self, tgt, flags, ttype)      : pass

    @SYNC
    def make_dir                (self, tgt, flags, ttype)      : pass
    @ASYNC
    def make_dir_async          (self, tgt, flags, ttype)      : pass


# ------------------------------------------------------------------------------





__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils            as ru
import radical.utils.signatures as rus


# ------------------------------------------------------------------------------
#
class Url (ru.Url) :
    """ 
    The SAGA Url class.

    URLs are used in several places in the SAGA API: to specify service
    endpoints for job submission or resource management, for file or
    directory locations, etc.

    The URL class is designed to simplify URL management for these
    purposes -- it allows to manipulate individual URL elements, while
    ensuring that the resulting URL is well formatted. Example::

      # create a URL from a string
      location = saga.Url ("file://localhost/tmp/file.dat")
      d = saga.filesystem.Directory(location)

    A URL consists of the following components (where one ore more can
    be 'None')::

      <scheme>://<user>:<pass>@<host>:<port>/<path>?<query>#<fragment>

    Each of these components can be accessed via its property or
    alternatively, via getter / setter methods. Example::

      url = saga.Url ("scheme://pass:user@host:123/path?query#fragment")

      # modify the scheme
      url.scheme = "anotherscheme"

      # above is equivalent with
      url.set_scheme("anotherscheme")
    """

    @rus.takes   ('Url', 
                  rus.optional (str, 'Url'))
    @rus.returns (rus.nothing)
    def __init__ (self, url_in=''):
        """ 
        __init__(url_string='')

        Create a new Url object from a string or another Url object.
        """

        self._super = super (Url, self)
        self._super.__init__ (url_in)


    # --------------------------------------------------------------------------
    #
    @rus.takes   ('Url', 
                 ('Url', dict))
    @rus.returns ('Url')
    def __deepcopy__ (self, memo) :
        """ 
        __deepcopy__(self, memo)

        Deep copy of a Url
        """

        return Url (self)


# ------------------------------------------------------------------------------
#






__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from ...exceptions import *
from ..            import base
from ..cpi         import SYNC_CALL, ASYNC_CALL
from ..cpi         import context as cpi
from ...           import context as api
from ...exceptions import *


######################################################################
#
# adaptor meta data
#
_ADAPTOR_NAME          = 'radical.saga.adaptors.userpass'
_ADAPTOR_SCHEMAS       = ['UserPass']
_ADAPTOR_OPTIONS       = []

_ADAPTOR_CAPABILITIES  = {
    'attributes'       : [api.TYPE,
                          api.USER_ID,
                          api.USER_PASS]
}

_ADAPTOR_DOC           = {
    'name'             : _ADAPTOR_NAME,
    'cfg_options'      : _ADAPTOR_OPTIONS, 
    'capabilities'     : _ADAPTOR_CAPABILITIES,
    'description'      : """This adaptor stores user_id and user_pass tokens, to
                            be used for backend connections.""",
    'schemas'          : {'userpass' : 'this adaptor can only store username/password pairs.'},
}

_ADAPTOR_INFO          = {
    'name'             : _ADAPTOR_NAME,
    'version'          : 'v0.1',
    'schemas'          : _ADAPTOR_SCHEMAS,
    'cpis'             : [{ 
        'type'         : 'radical.saga.Context',
        'class'        : 'ContextUserPass'
        }
    ]
}


###############################################################################
# The adaptor class

class Adaptor (base.Base):
    """ 
    This is the actual adaptor class, which gets loaded by SAGA (i.e. by the
    SAGA engine), and which registers the CPI implementation classes which
    provide the adaptor's functionality.
    """

    def __init__ (self) :

        base.Base.__init__ (self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        # there are no default myproxy contexts
        self._default_contexts = []


    def sanity_check (self) :
        pass


    def _get_default_contexts (self) :

        return self._default_contexts



######################################################################
#
# job adaptor class
#
class ContextUserPass (cpi.Context) :

    def __init__ (self, api, adaptor) :

        _cpi_base = super  (ContextUserPass, self)
        _cpi_base.__init__ (api, adaptor)


    @SYNC_CALL
    def init_instance (self, adaptor_state, type) :
        
        if not type.lower () in (schema.lower() for schema in _ADAPTOR_SCHEMAS) :
            raise BadParameter \
                    ("the UserPass context adaptor only handles UserPass contexts - duh!")

        self._type = type

        return self


    @SYNC_CALL
    def _initialize (self, session) :
        pass









__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


# FIXME: OS enums, ARCH enums

# resource type enum
COMPUTE      =  1;            """ resource accepting jobs """
STORAGE      =  2;            """ storage resource (duh) """
NETWORK      =  4;            """ connects compute and storage resources """

# resource state enum """
UNKNOWN      =  None;         """ wut? """
NEW          =  1;            """ requsted, not accepting jobs, yet;
                                  initial state """
PENDING      =  2;            """ accepting jobs, will become active eventually """
ACTIVE       =  4;            """ accepting jobs, jobs can run """
CANCELED     =  8;            """ released by user; final state """
EXPIRED      = 16;            """ released by system; final state """
DONE         = EXPIRED;       """ alias """
FAILED       = 32;            """ released unexpectedly by system or internally;
                                  final state """
FINAL        = CANCELED | DONE | FAILED

# resource attributes """
ID           = 'Id';          """ url identifying a resource instance """
RTYPE        = 'Rtype';       """ type enum, identifying the resource type """
STATE        = 'State';       """ state enum, identifying the rsource state """
STATE_DETAIL = 'StateDetail'; """ string, representing the native backend state """
MANAGER      = 'Manager';     """ url, pointing to the resource's manager """
DESCRIPTION  = 'Description'; """ dict, containing resource attributes  """

# generic resource description attributes """
TEMPLATE     = 'Template';    """ string, template to which the resource 
                                  was created"""
IMAGE        = 'Image';       """ FIXME: """

# resource lifetime attributes """
DYNAMIC      = 'Dynamic';     """ bool, enables/disables on-demand 
                                  resource 
                                  resizing """
START        = 'Start';       """ time, expected time at which resource 
                                  becomes ACTIVE """
END          = 'End';         """ time, expected time at which resource 
                                  will EXPIRE """
DURATION     = 'Duration';    """ time, expected time span between ACTIVE 
                                  and EXPIRED """

# resource type specific (non-generic) attributes """
MACHINE_OS   = 'MachineOS';   """ enum, identifying the resource's operating 
                                  system """
MACHINE_ARCH = 'MachineArch'; """ enum, identifying the machine architecture """
SIZE         = 'Size';        """ int, identifying No. of slots / size in 
                                  MB / No. of IPs """
MEMORY       = 'Memory';      """ int, identifying memory size in MegaByte """
ACCESS       = 'Access';      """ string, identifying the hostname/ip, mount
                                  point or provisioning URL to access the
                                  resource """







__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


from .manager   import Manager
from .resource  import Resource
from .resource  import Compute
from .resource  import Storage
from .resource  import Network

# ------------------------------------------------------------------------------





__author__    = "Mark Santcroos, Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2015, The RADICAL Project"
__license__   = "MIT"


""" TORQUE job adaptor implementation
"""

import re
import os
import time
import datetime
import threading

from urllib.parse import parse_qs

import radical.utils as ru

from ...              import exceptions as rse
from ...utils         import pty_shell  as rsups
from ...              import job        as api
from ...adaptors      import base       as a_base
from ...adaptors.cpi  import job        as cpi
from ...adaptors.cpi  import decorators as cpi_decs


SYNC_CALL  = cpi_decs.SYNC_CALL
ASYNC_CALL = cpi_decs.ASYNC_CALL


SYNC_WAIT_UPDATE_INTERVAL =  1  # seconds
MONITOR_UPDATE_INTERVAL   = 60  # seconds


# --------------------------------------------------------------------
#
class _job_state_monitor(threading.Thread):
    """ thread that periodically monitors job states
    """
    def __init__(self, job_service):

        self.logger = job_service._logger
        self.js = job_service
        self._term = threading.Event()

        super(_job_state_monitor, self).__init__()
        self.setDaemon(True)


    def stop(self):
        self._term.set()


    def run(self):

        # we stop the monitoring thread when we see the same error 3 times in
        # a row...
        error_type_count = dict()

        while not self._term.is_set ():

            try:
                # FIXME: do bulk updates here! we don't want to pull information
                # job by job. that would be too inefficient!
                jobs = self.js.jobs

                for job_id in list(jobs.keys()) :

                    job_info = jobs[job_id]

                    # we only need to monitor jobs that are not in a
                    # terminal state, so we can skip the ones that are
                    # either done, failed or canceled
                    if  job_info['state'] not in api.FINAL:

                        # Store the current state since the current state
                        # variable is updated when _job_get_info is called
                        pre_update_state = job_info['state']

                        new_job_info = self.js._job_get_info(job_id,
                                                             reconnect=False)
                        self.logger.info ("Job monitoring thread updating Job "
                                          "%s (old state: %s, new state: %s)" %
                                          (job_id, pre_update_state,
                                           new_job_info['state']))

                        # fire job state callback if 'state' has changed
                        if  new_job_info['state'] != pre_update_state:
                            job_obj = job_info['obj']
                            job_obj._attributes_i_set('state',
                                                      new_job_info['state'],
                                                      job_obj._UP, True)

                        # update job info
                        jobs[job_id] = new_job_info

            except Exception as e:
                import traceback
                traceback.print_exc ()
                self.logger.warning("Exception in job monitor thread: %s" % e)

                # check if we see the same error again and again
                error_type = str(e)
                if  error_type not in error_type_count :
                    error_type_count = dict()
                    error_type_count[error_type]  = 1
                else :
                    error_type_count[error_type] += 1
                    if  error_type_count[error_type] >= 3 :
                        self.logger.error("too many monitoring errors -- stop")
                        return

            finally :
                time.sleep (MONITOR_UPDATE_INTERVAL)


# --------------------------------------------------------------------
#
def _to_saga_jobstate(job_state, retcode, logger=None):
    """ translates one-letter batch system state to saga
    """
    # H: Job is held.
    # Q: Job is queued (eligible to run or routed.)
    # S: Job is suspended.
    # W: Job is waiting for its execution time to be reached.
    # R: Job is running.
    # E: Job is exiting after having run
    # T: Job is being moved to new location.
    # X: Subjob has completed execution or has been deleted (PBSPro)
    # F: Job is Finished (PBSPro)
    # C: Job is completed after having run (Torque)

    ret = None
    if   job_state == 'F':                 # PBSPro
        if retcode ==  0 : ret = api.DONE
        else             : ret = api.FAILED
    elif job_state == 'C':                 # Torque
        if retcode ==  0 : ret = api.DONE
        else             : ret = api.FAILED
    elif job_state == 'H': ret = api.PENDING
    elif job_state == 'Q': ret = api.PENDING
    elif job_state == 'S': ret = api.PENDING
    elif job_state == 'W': ret = api.PENDING
    elif job_state == 'R': ret = api.RUNNING
    elif job_state == 'E': ret = api.RUNNING
    elif job_state == 'T': ret = api.RUNNING
    elif job_state == 'X': ret = api.CANCELED  # PBSPro
    else                 : ret = api.UNKNOWN

    if logger:
        logger.debug('check state: %s', job_state)
        logger.debug('use   state: %s', ret)

    return ret


# --------------------------------------------------------------------
#
def _script_generator(url, logger, jd, ppn, gpn, gres, version,
                      is_cray=False, queue=None):
    '''
    generates a Torque script from a SAGA job description
    '''

    pbs_params  = str()
    exec_n_args = str()

    if jd.executable:
        exec_n_args += "%s " % (jd.executable)

    if jd.arguments:
        for arg in jd.arguments:
            exec_n_args += "%s " % (arg)

    if jd.processes_per_host:
        logger.info("Overriding detected ppn (%d) with processes_per_host (%d)"
                    % (ppn, jd.processes_per_host))
        ppn = jd.processes_per_host

    if jd.name:
        pbs_params += "#PBS -N %s \n" % jd.name

    if is_cray or 'Version: 4.2.7' not in version:
        # qsub on Cray systems complains about the -V option:
        # Warning:
        # Your job uses the -V option, which requests that all of your
        # current shell environment settings (9913 bytes) be exported to
        # it.  This is not recommended, as it causes problems for the
        # batch environment in some cases.
        pbs_params += "#PBS -V \n"

    if jd.environment:
        pbs_params += "#PBS -v %s\n" % \
                ','.join (["%s=%s" % (k,v)
                           for k,v in jd.environment.items()])

    # apparently this doesn't work with older PBS installations
    #
    #   if jd.working_directory:
    #       pbs_params += "#PBS -d %s \n" % jd.working_directory
    #
    # a workaround is to do an explicit 'cd'
    if jd.working_directory:
        workdir_directives  = 'export    PBS_O_WORKDIR=%s \n' \
                                               % jd.working_directory
        workdir_directives += 'mkdir -p  %s\n' % jd.working_directory
        workdir_directives += 'cd        %s\n' % jd.working_directory
    else:
        workdir_directives = ''

    if jd.output:
        # if working directory is set, we want stdout to end up in
        # the working directory as well, unless it containes a specific
        # path name.
        if jd.working_directory:
            if os.path.isabs(jd.output):
                pbs_params += "#PBS -o %s \n" % jd.output
            else:
                # user provided a relative path for STDOUT. in this case
                # we prepend the workind directory path before passing
                # it on to PBS
                pbs_params += "#PBS -o %s/%s \n" % (jd.working_directory,
                                                    jd.output)
        else:
            pbs_params += "#PBS -o %s \n" % jd.output

    if jd.error:
        # if working directory is set, we want stderr to end up in
        # the working directory as well, unless it contains a specific
        # path name.
        if jd.working_directory:
            if os.path.isabs(jd.error):
                pbs_params += "#PBS -e %s \n" % jd.error
            else:
                # user provided a realtive path for STDERR. in this case
                # we prepend the workind directory path before passing
                # it on to PBS
                pbs_params += "#PBS -e %s/%s \n" % (jd.working_directory,
                                                    jd.error)
        else:
            pbs_params += "#PBS -e %s \n" % jd.error


    if jd.wall_time_limit:
        hours = int(jd.wall_time_limit / 60)
        minutes = jd.wall_time_limit % 60
        pbs_params += "#PBS -l walltime=%s:%s:00 \n" \
            % (str(hours), str(minutes))

    # see https://gist.github.com/nobias/5b2373258e595e5242d5
    # The parameter to '-q' can have the following forms:
    #
    #   queue
    #   queue@server
    #   @server
    #
    # where 'server' is the target resource which can be *different* than the
    # submission host.  We interpret 'jd.candidate_hosts[0]' as such a target
    # resource - but only if exactly one `candidate_host` is given.

    # We haqve to take care to filter out special cases where we abise
    # `candidate_hosts` for node properties (those are appended to the nodes
    # argument in the resource_list).  This is currently only implemented for
    # "bigflash" on Gordon@SDSC
    #
    # https://github.com/radical-cybertools/radical.saga/issues/406

    queue_spec      = ''
    node_properties = []

    if      queue: queue_spec =    queue
    elif jd.queue: queue_spec = jd.queue

    # Parse candidate_hosts
    #
    # Currently only implemented for "bigflash" on Gordon@SDSC
    # https://github.com/radical-cybertools/radical.saga/issues/406
    #
    if jd.candidate_hosts:
        if 'BIG_FLASH' in jd.candidate_hosts:
            node_properties.append('bigflash')
        elif len(jd.candidate_hosts) == 1:
            queue_spec += '@%s' % jd.candidate_hosts[0]
        else:
            raise rse.NotImplemented("unsupported candidate_hosts [%s]"
                                % jd.candidate_hosts)
    if queue_spec:
        pbs_params += "#PBS -q %s\n" % queue_spec

    if jd.project:

        if '@' in jd.project:
            user, group = jd.project.split('@', 1)
            pbs_params += "#PBS -A %s \n"            % user
            pbs_params += "#PBS -W group_list:%s \n" % group

        elif 'PBSPro_1' in version:
            # On PBS Pro we set both -P(roject) and -A(accounting),
            # as we don't know what the admins decided, and just
            # pray that this doesn't create problems.
            pbs_params += "#PBS -P %s \n" % str(jd.project)
            pbs_params += "#PBS -A %s \n" % str(jd.project)

        else:
            # Torque
            pbs_params += "#PBS -A %s \n" % str(jd.project)

    if jd.job_contact:
        pbs_params += "#PBS -m abe \n"

    # if total_cpu_count is not defined, we assume 0
    if not jd.total_cpu_count:
        jd.total_cpu_count = 0

    # if total_gpu_count is not defined, we assume 0
    if not jd.total_gpu_count:
        jd.total_gpu_count = 0

    # make sure we want CPUs or GPUs
    assert(jd.total_cpu_count + jd.total_gpu_count > 0)

    # Request enough nodes to cater for the number of cores requested
    cpu_nnodes = 0
    if jd.total_cpu_count:
        cpu_nnodes = jd.total_cpu_count / ppn
        if jd.total_cpu_count % ppn > 0:
            cpu_nnodes += 1

    # Request enough nodes to cater for the number of cores requested
    gpu_nnodes = 0
    if jd.total_gpu_count:
        gpu_nnodes = jd.total_gpu_count / gpn
        if jd.total_gpu_count % gpn > 0:
            gpu_nnodes += 1

    nnodes = max([cpu_nnodes, gpu_nnodes])

    # use the ncpus for systems that need to specify ncpus as multiple of PPN
    ncpus = nnodes * ppn

    if is_cray:
        # Special cases for PBS/TORQUE on Cray. Different PBSes,
        # different flags. A complete nightmare...
        if  'titan'      in url.host       or \
            '5.1.0.h1'   in version or \
            '6.1.1.1.h2' in version or \
            '4.2.6-snap' in version :
            # Can't really use hostname as we run also from the headnode
            logger.info("Using Titan (Cray XP) specific '#PBS -l nodes=xx'")
            pbs_params += "#PBS -l nodes=%d\n" % nnodes
        elif 'edison' in url.host:

            logger.info("Using Edison@NERSC (Cray XC30) (mppwidth=xx)")
            pbs_params += "#PBS -l mppwidth=%s \n" % jd.total_cpu_count
        elif 'bw.ncsa.illinois.edu' in url.host   or \
             'Version: 6.0.4'       in version:
            if gpu_nnodes: gpu_flag = ':xk'
            else         : gpu_flag = ''
            if gpu_flag and ppn > 16: ppn = 16
            logger.info("Using Blue Waters (Cray XE6/XK7) (nodes=xx:ppn=yy)")
            pbs_params += "#PBS -l nodes=%d:ppn=%d%s\n" \
                        % (nnodes, ppn, gpu_flag)
            if 'RS_USE_COMMTRANSPARENT' in os.environ:
                pbs_params += "#PBS -l flags=commtransparent\n"

        elif 'Version: 5.' in version:
            # What would removing this catchall break?
            logger.info("Using TORQUE 5.x notation '#PBS -l procs=XX' ")
            pbs_params += "#PBS -l procs=%d\n" % jd.total_cpu_count
        else:
            logger.info("Using Cray XT (e.g. Kraken, Jaguar) (size=xx)")
            pbs_params += "#PBS -l size=%s\n" % jd.total_cpu_count
    elif 'version: 2.3.13' in version:
        # e.g. Blacklight
        # TODO: The more we add, the more it screams for a refactoring
        pbs_params += "#PBS -l ncpus=%d\n" % ncpus
    elif 'hopper' in url.host:
        logger.info("Using Hopper@NERSC (Cray XE6) (mppwidth=xx)")
        pbs_params += "#PBS -l mppwidth=%s \n" % jd.total_cpu_count
    elif 'rhea.ccs.ornl.gov' in url.host or 'rhea-login' in os.uname()[1]:
        # Not allowed to specify ppn on Rhea
        pbs_params += "#PBS -l nodes=%d\n" % (nnodes)
    else:
        # Default case, i.e, standard HPC cluster (non-Cray)

        # If we want just a slice of one node
        if jd.total_cpu_count < ppn:
            ppn = jd.total_cpu_count

      # pbs_params += "#PBS -l nodes=%d:ppn=%d%s\n" % (
        pbs_params += "#PBS -l nodes=%d%s\n" % (
            nnodes, ''.join([':%s' % prop for prop in node_properties]))

    # Process Generic Resource specification request
    if gres:
        pbs_params += "#PBS -l gres=%s\n" % gres

    exec_n_args += 'export SAGA_PPN=%d\n' % ppn
    exec_n_args += 'export SAGA_GPN=%d\n' % gpn

    # escape all double quotes and dollarsigns, otherwise 'echo |'
    # further down won't work
    # only escape '$' in args and exe. not in the params
    exec_n_args = workdir_directives + exec_n_args
    exec_n_args = exec_n_args.replace('$', '\\$')

    pbscript = "\n#!/bin/bash \n%s%s" % (pbs_params, exec_n_args)

    pbscript = pbscript.replace('"', '\\"')
    return pbscript


# --------------------------------------------------------------------
# some private defs
#
_PTY_TIMEOUT = 2.0

# --------------------------------------------------------------------
# the adaptor name
#
_ADAPTOR_NAME          = "radical.saga.adaptors.torquejob"
_ADAPTOR_SCHEMAS       = ["torque", "torque+ssh", "torque+gsissh"]
_ADAPTOR_OPTIONS       = []

# --------------------------------------------------------------------
# the adaptor capabilities & supported attributes
#
_ADAPTOR_CAPABILITIES = {
    "jdes_attributes":   [api.NAME,
                          api.EXECUTABLE,
                          api.ARGUMENTS,
                          api.CANDIDATE_HOSTS,
                          api.ENVIRONMENT,
                          api.INPUT,
                          api.OUTPUT,
                          api.ERROR,
                          api.QUEUE,
                          api.PROJECT,
                          api.FILE_TRANSFER,
                          api.WALL_TIME_LIMIT,
                          api.WORKING_DIRECTORY,
                          api.WALL_TIME_LIMIT,
                          api.PROCESSES_PER_HOST,
                          api.SPMD_VARIATION,
                          api.TOTAL_CPU_COUNT,
                          api.TOTAL_GPU_COUNT,
                         ],
    "job_attributes":    [api.EXIT_CODE,
                          api.EXECUTION_HOSTS,
                          api.CREATED,
                          api.STARTED,
                          api.FINISHED],
    "metrics":           [api.STATE],
    "callbacks":         [api.STATE],
    "contexts":          {"ssh": "SSH public/private keypair",
                          "x509": "GSISSH X509 proxy context",
                          "userpass": "username/password pair (ssh)"}
}

# --------------------------------------------------------------------
# the adaptor documentation
#
_ADAPTOR_DOC = {
    "name":          _ADAPTOR_NAME,
    "cfg_options":   _ADAPTOR_OPTIONS,
    "capabilities":  _ADAPTOR_CAPABILITIES,
    "description":  """
The TORQUE adaptor allows to run and manage jobs on
`TORQUE <http://www.adaptivecomputing.com/products/open-source/torque>`_
controlled HPC clusters.
""",
    "example": "examples/jobs/torquejob.py",
    "schemas": {"torque":        "connect to a local cluster",
                "torque+ssh":    "connect to a remote cluster via SSH",
                "torque+gsissh": "connect to a remote cluster via GSISSH"}
}

# --------------------------------------------------------------------
# the adaptor info is used to register the adaptor with SAGA
#
_ADAPTOR_INFO = {
    "name"        : _ADAPTOR_NAME,
    "version"     : "v0.1",
    "schemas"     : _ADAPTOR_SCHEMAS,
    "capabilities": _ADAPTOR_CAPABILITIES,
    "cpis": [
                {
                    "type": "radical.saga.job.Service",
                    "class": "TORQUEJobService"
                },
                {
                    "type": "radical.saga.job.Job",
                    "class": "TORQUEJob"
                }
            ]
}


###############################################################################
# The adaptor class
class Adaptor (a_base.Base):
    """ this is the actual adaptor class, which gets loaded by SAGA (i.e. by
        the SAGA engine), and which registers the CPI implementation classes
        which provide the adaptor's functionality.
    """

    # ----------------------------------------------------------------
    #
    def __init__(self):

        a_base.Base.__init__(self, _ADAPTOR_INFO, _ADAPTOR_OPTIONS)

        self.id_re = re.compile('^\[(.*)\]-\[(.*?)\]$')
        self.epoch = datetime.datetime(1970, 1, 1)


    # ----------------------------------------------------------------
    #
    def sanity_check(self):
        # FIXME: also check for gsissh
        pass


    # ----------------------------------------------------------------
    #
    def parse_id(self, id):
        # split the id '[rm]-[pid]' in its parts, and return them.

        match = self.id_re.match(id)

        if not match or len(match.groups()) != 2:
            raise rse.BadParameter("Cannot parse job id '%s'" % id)

        return (match.group(1), match.group(2))


###############################################################################
#
class TORQUEJobService (cpi.Service):
    """ implements cpi.Service
    """

    # ----------------------------------------------------------------
    #
    def __init__(self, api, adaptor):

        self._mt  = None
        _cpi_base = super(TORQUEJobService, self)
        _cpi_base.__init__(api, adaptor)

        self._adaptor = adaptor

    # ----------------------------------------------------------------
    #
    def __del__(self):

        self.close()


    # ----------------------------------------------------------------
    #
    def close(self):

        if  self.mt :
            self.mt.stop()
            self.mt.join(10)  # don't block forever on join()

        self._logger.info("Job monitoring thread stopped.")
        self.finalize(True)


    # ----------------------------------------------------------------
    #
    def finalize(self, kill_shell=False):

        if  kill_shell :
            if  self.shell :
                self.shell.finalize (True)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def init_instance(self, adaptor_state, rm_url, session):
        """ service instance constructor
        """
        self.rm      = rm_url
        self.session = session
        self.ppn     = None
        self.gpn     = 1   # gpus per node  # FIXME: inspect system
        self.is_cray = False
        self.queue   = None
        self.shell   = None
        self.jobs    = dict()
        self.gres    = None

        # the monitoring thread - one per service instance
        self.mt = _job_state_monitor(job_service=self)
        self.mt.start()

        rm_scheme = rm_url.scheme
        pty_url   = ru.Url(rm_url)

        # this adaptor supports options that can be passed via the
        # 'query' component of the job service URL.
        if rm_url.query:
            for key, val in parse_qs(rm_url.query).items():
                if key == 'queue':
                    self.queue = val[0]
                elif key == 'craytype':
                    self.is_cray = val[0]
                elif key == 'ppn':
                    self.ppn = int(val[0])
                elif key == 'gres':
                    self.gres = val[0]


        # we need to extract the scheme for PTYShell. That's basically the
        # job.Service Url without the pbs+ part. We use the PTYShell to execute
        # pbs commands either locally or via gsissh or ssh.
        scheme_elems = rm_scheme.split('+')
        if   'gsissh' in scheme_elems : pty_url.scheme = "gsissh"
        elif 'ssh'    in scheme_elems : pty_url.scheme = "ssh"
        else                          : pty_url.scheme = "fork"

        # these are the commands that we need in order to interact with PBS.
        # the adaptor will try to find them during initialize(self) and bail
        # out in case they are note available.
        self._commands = {'pbsnodes': None,
                          'qstat':    None,
                          'qsub':     None,
                          'qdel':     None,
                          'checkjob': None}

        self.shell = rsups.PTYShell(pty_url, self.session)

      # self.shell.set_initialize_hook(self.initialize)
      # self.shell.set_finalize_hook(self.finalize)

        self.initialize()
        return self.get_api()


    # ----------------------------------------------------------------
    #
    def initialize(self):

        # check if all required pbs tools are available
        ret, out, _ = self.shell.run_sync("qstat --version")

        if ret:
            raise rse.NoSuccess("Error finding PBS tools: %s" % out)

        self._pbs_version = out.strip()
        self._logger.info("Found PBS version: %s" % self._pbs_version)

        for cmd in self._commands:

            ret, out, _ = self.shell.run_sync("which %s " % cmd)

            if ret:
                self._logger.warn("missing PBS tool %s: %s" % (cmd, out))
                continue

            self._commands[cmd] = out.strip()  # strip removes newline
            self._logger.info("Found PBS %s: %s" % (cmd, out.strip()))


        # TODO: Get rid of this, as I dont think there is any justification that
        #       Cray's are special
        #
        # let's try to figure out if we're working on a Cray machine.
        # naively, we assume that if we can find the 'aprun' command in the
        # path that we're logged in to a Cray machine.
        if not self.is_cray:
            ret, out, _ = self.shell.run_sync('which aprun')
            if ret != 0:
                self.is_cray = False
            else:
                self._logger.info("Host '%s' seems to be a Cray machine."
                    % self.rm.host)
                self.is_cray = True
        else:
            self._logger.info("host is cray: %s" % self.is_cray)

        #
        # Get number of processes per node
        #
        if self.ppn:
            self._logger.debug("Using user specified 'ppn': %d" % self.ppn)
            return

        # TODO: this is quite a hack. however, it *seems* to work quite
        #       well in practice.
        if any(ver in  self._pbs_version
                   for ver in ('PBSPro_13', 'PBSPro_12', 'PBSPro_11.3')):
            ret, out, _ = self.shell.run_sync('unset GREP_OPTIONS; %s -a '
                          '| grep -E "resources_available.ncpus"'
                          % self._commands['pbsnodes'])
        else:
            ret, out, _ = self.shell.run_sync('unset GREP_OPTIONS; %s -a '
                          '| grep -E "(np|pcpu|pcpus)[[:blank:]]*=" '
                          % self._commands['pbsnodes'])

        if ret != 0:
            raise rse.NoSuccess("Error running pbsnodes: %s" % out)
        else:
            # this is black magic. we just assume that the highest occurrence
            # of a specific np is the number of processors (cores) per compute
            # node. this equals max "PPN" for job scripts
            ppn_list = dict()
            for line in out.split('\n'):
                np = line.split(' = ')
                if len(np) == 2:
                    np_str = np[1].strip()
                    if np_str == '<various>':
                        continue
                    else:
                        np = int(np_str)
                    if np in ppn_list:
                        ppn_list[np] += 1
                    else:
                        ppn_list[np] = 1
            self.ppn = max(ppn_list, key=ppn_list.get)
            self._logger.debug("Found the following 'ppn' configurations: %s. "
                "Using %s as default ppn."  % (ppn_list, self.ppn))

    # ----------------------------------------------------------------
    #
    def _job_run(self, job_obj):
        """ runs a job via qsub
        """

        # get the job description
        jd       = job_obj.get_description()
        job_name = jd.name

        # normalize working directory path
        if  jd.working_directory :
            jd.working_directory = os.path.normpath (jd.working_directory)

        # TODO: Why would one want this?
        if self.queue and jd.queue:
            self._logger.warning("Job service was instantiated explicitly with \
'queue=%s', but job description tries to a different queue: '%s'. Using '%s'." %
                                (self.queue, jd.queue, self.queue))

        try:
            # create a PBS job script from SAGA job description
            script = _script_generator(url=self.rm, logger=self._logger,
                                   jd=jd, ppn=self.ppn, gpn=self.gpn,
                                   gres=self.gres,
                                   version=self._pbs_version,
                                   is_cray=self.is_cray, queue=self.queue
                                   )

            self._logger.info("Generated PBS script: %s" % script)
        except Exception as e:
            raise rse.BadParameter('error generating pbs script') from e

        # try to create the working directory (if defined)
        # WARNING: this assumes a shared filesystem between login node and
        #          compute nodes.
        if jd.working_directory:
            self._logger.info("Create workdir %s" % jd.working_directory)
            ret, out, _ = self.shell.run_sync("mkdir -p %s"
                                             % jd.working_directory)
            if ret != 0:
                # something went wrong
                raise rse.NoSuccess("Couldn't create workdir - %s" % out)

        # Now we want to execute the script. This process consists of two steps:
        # (1) we create a temporary file with 'mktemp' and write the contents of
        #     the generated PBS script into it
        # (2) we call 'qsub <tmpfile>' to submit the script to the
        #     queueing system
        cmdline = """
        SCRIPTFILE=`mktemp -t rs.jobscript.XXXXXX` \\
            &&  echo "%s" > $SCRIPTFILE \\
            &&  %s $SCRIPTFILE \\
            &&  rm -f $SCRIPTFILE
            """ %  (script, self._commands['qsub'])
        ret, out, _ = self.shell.run_sync(cmdline)

        if ret != 0:
            # something went wrong
            raise rse.NoSuccess("Error running qsub: %s: %s" % (out, cmdline))
        else:
            # parse the job id. qsub usually returns just the job id, but
            # sometimes there are a couple of lines of warnings before.
            # if that's the case, we log those as 'warnings'
            lines = out.split('\n')
            lines = [lines for lines in lines if lines != '']  # remove empty

            if len(lines) > 1:
                self._logger.warning('qsub: %s' % ''.join(lines[:-2]))

            # we asssume job id is in the last line
            job_id = "[%s]-[%s]" % (self.rm, lines[-1].strip().split('.')[0])
            self._logger.info("Submitted PBS job with id: %s" % job_id)

            state = api.PENDING

            # populate job info dict
            self.jobs[job_id] = {'obj'         : job_obj,
                                 'job_id'      : job_id,
                                 'name'        : job_name,
                                 'state'       : state,
                                 'exec_hosts'  : None,
                                 'returncode'  : None,
                                 'create_time' : time.time(),
                                 'start_time'  : None,
                                 'end_time'    : None,
                                 'gone'        : False
                                 }

            self._logger.info ("assign job id  %s / %s / %s to watch list (%s)"
                            % (job_name, job_id, job_obj, list(self.jobs.keys())))

            # set status to 'pending' and manually trigger callback
            job_obj._attributes_i_set('state', state, job_obj._UP, True)

            # return the job id
            return job_id


    # ----------------------------------------------------------------
    #
    def _job_get_info(self, job_id, reconnect):
        """ Get job information attributes via qstat.
        """

        # If we don't have the job in our dictionary, we don't want it,
        # unless we are trying to reconnect.
        if not reconnect and job_id not in self.jobs:
            raise rse.NoSuccess("Unknown job id: %s" % job_id)

        if not reconnect:
            # job_info contains the info collect when _job_get_info
            # was called the last time
            job_info = self.jobs[job_id]

            # if the 'gone' flag is set, there's no need to query the job
            # state again. it's gone forever
            if job_info['gone'] is True:
                return job_info
        else:
            # Create a template data structure
            job_info = {
                'job_id':       job_id,
                'state':        api.UNKNOWN,
                'name':         None,
                'exec_hosts':   None,
                'returncode':   None,
                'create_time':  None,
                'start_time':   None,
                'end_time':     None,
                'gone':         False
            }

        rm, pid = self._adaptor.parse_id(job_id)

        # try `qstat`.  If that doesn't work, fall back to `checkjob`
        ok = False

        # TODO: move to config file
        #       if 'PBSPro_1' in self._pbs_version:
        qstat_flag = '-f1'

        ret, out1, err1 = self.shell.run_sync("unset GREP_OPTIONS; %s %s %s | "
                "grep -E -i '(job_state)|(exec_host)|(exit_status)|"
                 "(ctime)|(start_time)|(stime)|(mtime)'"
                % (self._commands['qstat'], qstat_flag, pid))

        if ret != 0:
            self._logger.warn('qstat failed with: %s', err1)

        else:

            # qstat worked - parse output
            ok = True

            # the result should look something like this:
            #     job_state = C
            #     exec_host = i72/0
            #     exit_status = 0
            for line in out1.split('\n'):

                if '=' not in line:
                    continue

                key, val = line.split('=', 1)
                key = key.strip().lower
                val = val.strip()

                if   key in ['job_state'  ]: job_state = val
                elif key in ['job_name'   ]: job_info['name'] = val
                elif key in ['exit_status']: job_info['returncode' ] = int(val)
                elif key in ['exec_host'  ]: job_info['exec_hosts' ] = val.split('+')
                                             # format i73/7+i73/6+...

              # FIXME: qstat will not tell us time zones, so we cannot
              #        convert to EPOCH (which is UTC).  We thus take
              #        times ourself.  A proper solution would be to
              #        either do the time conversion on the target host,
              #        or to inspect time zone settings on the host.
              #
              # # PBS Pro doesn't provide "end time", but
              # # "resources_used.walltime" could be added up to the
              # # start time.  Alternatively, we can use mtime, (latest
              # # modification time) which is generally also end time.
              # # TORQUE has an "comp_time" (completion? time), that is
              # # generally the same as mtime.  # # For now we  use
              # mtime for both TORQUE and PBS Pro.

                elif key in ['start_time',  # TORQUE / PBS Pro
                             'stime'      ]: job_info['start_time' ] = val
                elif key in ['ctime'      ]: job_info['create_time'] = val
                elif key in ['mtime'      ]: job_info['end_time'   ] = val


        if not ok:

            # qstat failed - try `checkjob`
            ret, out2, err2 = self.shell.run_sync("%s %s"
                    % (self._commands['checkjob'], pid))

            if ret != 0:
                self._logger.warn('checkjob failed with: %s', err2)

            else:

                # checkjob worked - parse result
                ok = True

                # the result should look something like this:
                #     job 4683422
                #
                #     AName: titan.qsub
                #     State: Running
                #     Creds:  user:merzky1  group:merzky1  account:BIP149 \
                #             class:debug  qos:smallplcmnt
                #     WallTime:   00:00:52 of 00:05:00
                #     SubmitTime: Mon Apr 29 06:38:53
                #     ...
                #
                # See also parse comments on qstat values
                # FIXME: exec host parsing
                # FIXME: missing start time (should be computed from walltime)
                for line in out2.split('\n'):

                    if not ':' in line:
                        continue

                    key, val = line.split(':', 1)
                    key = key.strip().lower()
                    val = val.strip()

                    if   key in ['state'     ]: job_state               = val
                    elif key in ['aname'     ]: job_info['name']        = val
                    elif key in ['submittime']: job_info['create_time'] = val

                    elif key in ['completion code']:
                        job_info['end_time'  ] = str(val.split(':', 1)[1])
                        job_info['returncode'] = int(val.split(      )[0])

        if ok:

            # we did get some information - see if we need a state update
            # TORQUE doesn't allow us to distinguish DONE/FAILED on
            # final state alone,  we need to consider the exit_status.
            retcode = job_info.get('returncode', -1)
            job_info['state'] = _to_saga_jobstate(job_state, retcode,
                                                  logger=self._logger)

            # FIXME: workaround for time zone problem described above
            if job_info['state'] in [api.RUNNING] + api.FINAL \
                and not job_info['start_time']:
                job_info['start_time'] = time.time()

            if job_info['state'] in api.FINAL \
                and not job_info['end_time']:
                job_info['end_time'] = time.time()


        else:

            # both failed - search all output for 'unknown job' or 'invalid job'
            all_out = '\n'.join([out1, out2, err1, err2])

            if  "unknown job" in all_out.lower() or \
                "invalid job" in all_out.lower():

                # Let's see if the last known job state was running or pending.
                # in that case, the job is gone now, which can either mean DONE,
                # or FAILED. the only thing we can do is set it to 'DONE'
                job_info['gone'] = True

                # TODO: we can also set the end time?
                self._logger.warning("Previously running job has disappeared. "
                    "This probably means that the backend doesn't store "
                    "information about finished jobs. Setting state to 'DONE'.")

                if job_info['state'] in [api.RUNNING, api.PENDING]:
                    job_info['state']      = api.DONE
                    job_info['returncode'] = 0  # we are guessing here...
                else:
                    job_info['state'] = api.UNKNOWN

                if not job_info['start_time']:
                    # inaccurate guess, but better than nothing
                    job_info['start_time'] = time.time()

                if not job_info['end_time']:
                    # inaccurate guess, but better than nothing
                    job_info['end_time'] = time.time()

            else:
                if reconnect:
                    raise rse.NoSuccess("Couldn't reconnect to '%s'" % job_id)

                else:
                    raise rse.NoSuccess("Error retrieving info for %s" % job_id)

        # return the updated job info
        return job_info


    # ----------------------------------------------------------------
    #
    def _parse_qstat(self, haystack, job_info):


        # return the new job info dict
        return job_info

    # ----------------------------------------------------------------
    #
    def _job_get_state(self, job_id):
        """ get the job's state
        """
        return self.jobs[job_id]['state']

    # ----------------------------------------------------------------
    #
    def _job_get_exit_code(self, job_id):
        """ get the job's exit code
        """
        ret = self.jobs[job_id]['returncode']

        if ret is None : return None
        else           : return int(ret)

    # ----------------------------------------------------------------
    #
    def _job_get_execution_hosts(self, job_id):
        """ get the job's exit code
        """
        return self.jobs[job_id]['exec_hosts']

    # ----------------------------------------------------------------
    #
    def _job_get_create_time(self, job_id):
        """ get the job's creation time
        """
        return self.jobs[job_id]['create_time']

    # ----------------------------------------------------------------
    #
    def _job_get_start_time(self, job_id):
        """ get the job's start time
        """
        return self.jobs[job_id]['start_time']

    # ----------------------------------------------------------------
    #
    def _job_get_end_time(self, job_id):
        """ get the job's end time
        """
        return self.jobs[job_id]['end_time']

    # ----------------------------------------------------------------
    #
    def _job_cancel(self, job_id):
        """ cancel the job via 'qdel'
        """
        rm, pid = self._adaptor.parse_id(job_id)

        ret, out, _ = self.shell.run_sync("%s %s\n"
                                       % (self._commands['qdel'], pid))

        if ret != 0:
            raise rse.NoSuccess("Error canceling job via 'qdel': %s" % out)

        # assume the job was succesfully canceled
        self.jobs[job_id]['state'] = api.CANCELED

        if not self.jobs[job_id]['end_time']:
            self.jobs[job_id]['end_time'] = time.time()


    # ----------------------------------------------------------------
    #
    def _job_wait(self, job_id, timeout):
        """ wait for the job to finish or fail
        """
        time_start = time.time()
        time_now   = time_start
        rm, pid    = self._adaptor.parse_id(job_id)

        while True:
            state = self.jobs[job_id]['state']  # this gets updated in the bg.

            if state in api.FINAL:
                return True

            # avoid busy poll
            time.sleep(SYNC_WAIT_UPDATE_INTERVAL)

            # check if we hit timeout
            if timeout >= 0:
                time_now = time.time()
                if time_now - time_start > timeout:
                    return False

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def create_job(self, jd):
        """ implements cpi.Service.get_url()
        """
        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       False
                         }

        # create and return a new job object
        return api.Job(_adaptor=self._adaptor,
                       _adaptor_state=adaptor_state)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_job(self, job_id):
        """ Implements cpi.Service.get_job()

            Re-create job instance from a job-id.
        """

        # If we already have the job info, we just pass the current info.
        if job_id in self.jobs :
            return self.jobs[job_id]['obj']

        # Try to get some initial information about this job (again)
        job_info = self._job_get_info(job_id, reconnect=True)

        # create job description from queue info
        # FIXME: incomplete
        jd      = api.Description()
        jd.name = job_info['name']

        # this dict is passed on to the job adaptor class -- use it to pass any
        # state information you need there.
        adaptor_state = {"job_service":     self,
                         "job_description": jd,
                         "job_schema":      self.rm.schema,
                         "reconnect":       True,
                         "reconnect_jobid": job_id
                         }

        job_obj = api.Job(_adaptor=self._adaptor,
                          _adaptor_state=adaptor_state)

        # throw it into our job dictionary.
        job_info['obj']   = job_obj
        self.jobs[job_id] = job_info

        return job_obj

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_url(self):
        """ implements cpi.Service.get_url()
        """
        return self.rm

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def list(self):
        """ implements cpi.Service.list()
        """
        ids = []

        ret, out, _ = self.shell.run_sync(
                                     "unset GREP_OPTIONS; %s | grep `whoami`"
                                     % self._commands['qstat'])

        if ret != 0 and len(out) > 0:
            raise rse.NoSuccess("failed to list jobs via 'qstat': %s" % out)
        elif ret != 0 and len(out) == 0:
            # qstat | grep `` exits with 1 if the list is empty
            pass
        else:
            for line in out.split("\n"):
                # output looks like this:
                # 112059.svc.uc.futuregrid testjob oweidner 0 Q batch
                # 112061.svc.uc.futuregrid testjob oweidner 0 Q batch
                if len(line.split()) > 1:
                    job_id = "[%s]-[%s]" % (self.rm,
                                            line.split()[0].split('.')[0])
                    ids.append(str(job_id))

        return ids


    # ----------------------------------------------------------------
    #
    def container_run (self, jobs) :

        self._logger.debug ("container run: %s"  %  str(jobs))

        # TODO: this is not optimized yet
        for job in jobs:
            job.run ()


    # ----------------------------------------------------------------
    #
    def container_wait (self, jobs, mode, timeout) :

        self._logger.debug ("container wait: %s"  %  str(jobs))

        # TODO: this is not optimized yet
        for job in jobs:
            job.wait ()


    # ----------------------------------------------------------------
    #
    def container_cancel (self, jobs, timeout) :

        self._logger.debug ("container cancel: %s"  %  str(jobs))

        # TODO: this is not optimized yet
        for job in jobs:
            job.cancel (timeout)


################################################################################
#
class TORQUEJob (cpi.Job):
    """ implements cpi.Job
    """


    def __init__(self, api, adaptor):

        # initialize parent class
        _cpi_base = super(TORQUEJob, self)
        _cpi_base.__init__(api, adaptor)


    def _get_impl(self):
        return self


    @SYNC_CALL
    def init_instance(self, job_info):
        """ implements cpi.Job.init_instance()
        """

        # init_instance is called for every new api.Job object
        # that is created
        self.jd = job_info["job_description"]
        self.js = job_info["job_service"]

        if job_info['reconnect'] is True:
            self._id      = job_info['reconnect_jobid']
            self._name    = self.jd.name
            self._started = True
        else:
            self._id      = None
            self._name    = self.jd.name
            self._started = False

        return self.get_api()


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_state(self):
        """ implements cpi.Job.get_state()
        """

        if  self._started is False:
            return api.NEW

        return self.js._job_get_state(job_id=self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def wait(self, timeout):
        """ implements cpi.Job.wait()
        """

        if self._started is False:
            raise rse.IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_wait(job_id=self._id, timeout=timeout)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def cancel(self, timeout):
        """ implements cpi.Job.cancel()
        """

        if self._started is False:
            raise rse.IncorrectState("Can't wait for job that hasn't been started")
        else:
            self.js._job_cancel(self._id)


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def run(self):
        """ implements cpi.Job.run()
        """

        self._id = self.js._job_run(self._api())
        self._started = True


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_service_url(self):
        """ implements cpi.Job.get_service_url()
        """

        return self.js.rm


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_id(self):
        """ implements cpi.Job.get_id()
        """

        return self._id


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_name (self):
        """ Implements cpi.Job.get_name() """

        return self._name


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_exit_code(self):
        """ implements cpi.Job.get_exit_code()
        """
        if self._started: return self.js._job_get_exit_code(self._id)
        else            : return None


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_created(self):
        """ implements cpi.Job.get_created()
        """
        if self._started: return self.js._job_get_create_time(self._id)
        else            : return None


    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_started(self):
        """ implements cpi.Job.get_started()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_start_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_finished(self):
        """ implements cpi.Job.get_finished()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_end_time(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_execution_hosts(self):
        """ implements cpi.Job.get_execution_hosts()
        """
        if self._started is False:
            return None
        else:
            return self.js._job_get_execution_hosts(self._id)

    # ----------------------------------------------------------------
    #
    @SYNC_CALL
    def get_description(self):
        """ implements cpi.Job.get_execution_hosts()
        """
        return self.jd


# ------------------------------------------------------------------------------

